"""
Defines all Django settings for Simmate.

Note, some settings are dynamically determined, such as the installed apps and
database backend being used.
"""

import os

from simmate.configuration import settings
from simmate.utilities import get_directory

# --------------------------------------------------------------------------------------

DEBUG = settings.website.debug

ALLOWED_HOSTS = settings.website.allowed_hosts

CSRF_TRUSTED_ORIGINS = settings.website.csrf_trusted_origins

SECRET_KEY = settings.website.secret_key

# --------------------------------------------------------------------------------------

# DATABASE CONNECTION
_default_database = {k.upper(): v for k, v in settings.database.items()}
DATABASES = {"default": _default_database}

# List all applications that Django will initialize with. Write the full python
# path to the app or it's config file. Note that relative python paths work too
# if you are developing a new app.
INSTALLED_APPS = [
    #
    # These are built-in django apps that we use for extra features
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "django.contrib.humanize",
    # Other apps installed with Django to consider
    #   "django.contrib.postgres",
    #   "django.contrib.redirects",
    #   "django.contrib.sitemaps",
    #
    # These are apps created by third-parties that give us extra features
    "crispy_forms",  # django-crispy-forms for HTML boostrap forms
    "rest_framework",  # djangorestframework for the REST API
    "django_filters",  # django-filter for filterable REST API urls
    #
    # Apps for django-allauth that allow sign-on using external accounts
    "allauth",
    "allauth.account",
    "allauth.socialaccount",
    # note: there are extra apps installed based on configuration. See the
    # allauth section at the bottom of this page for more.
    #
    # Django unicorn acts as a frontend framework for making dyanmic webpages
    # (i.e. AJAX calls can be made to update the views)
    "django_unicorn",
    #
    # Django simple history lets you track history of changes (and who made
    # those changes) for a given model. This is important for models that users
    # interact with and edit in the UI
    "simple_history",
    #
    # Django contrib comments let you track comments on a model and handles
    # moderation / user / date features for you.
    "django_comments",
    #
    # Other third-party apps/tools to consider. Note that some of these don't
    # need to be installed apps while some also request different setups.
    #   django-ratelimit
    #   dj-stripe
    #   django-graphene (+ GraphQL)
    #   django-redis
    #
    # These third-party tools are only to help with development and we should
    # try to avoid adding these as installed apps by default.
    #   "django_extensions",  # for development tools
    #   "debug_toolbar",  # django-debug-toolbar  # for debuging and profile-time info
    #
    # Simmate apps + user apps
    "simmate.website.configs.CoreComponentsConfig",
    "simmate.website.configs.DataExplorerConfig",
    "simmate.website.configs.WorkflowsConfig",
    "simmate.website.configs.WorkflowEngineConfig",
    *settings.apps,
]
# OPTIMIZE: for now, I just load everything, but I will want to allow users to
# to disable some apps from loading. This will allow faster import/setup times
# when running small scripts. One idea is to have an applications_override.yaml
# where the user specifies only what they want. Alternatively, I can use
# a general DATABASE_ONLY keyword or something similar to limit what's loaded.

# --------------------------------------------------------------------------------------

# EMAIL SETTINGS
EMAIL_BACKEND = settings.website.email.backend
EMAIL_HOST = settings.website.email.host
EMAIL_PORT = settings.website.email.port
EMAIL_USE_TLS = settings.website.email.use_tls
EMAIL_HOST_USER = settings.website.email.host_user
EMAIL_HOST_PASSWORD = settings.website.email.host_password
DEFAULT_FROM_EMAIL = settings.website.email.from_email
EMAIL_SUBJECT_PREFIX = settings.website.email.subject_prefix
EMAIL_TIMEOUT = settings.website.email.timeout
ADMINS = settings.website.admins

# --------------------------------------------------------------------------------------

# This sets the default field for primary keys in django models
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    # This is for tracking WHO changes a model (attaches User)
    "simple_history.middleware.HistoryRequestMiddleware",
    # adds specific authentication methods, such as login by email
    "allauth.account.middleware.AccountMiddleware",
]

# "core" here is based on the name of my main django folder
ROOT_URLCONF = "simmate.website.core.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            settings.config_directory / "templates",  # let's user easily override html
            settings.django_directory / "templates",  # a single templates folder
            # Then APP_DIRS are checked in order
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# "core" here is based on the name of my main django folder
WSGI_APPLICATION = "simmate.website.core.wsgi.application"

# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

DATETIME_INPUT_FORMATS = [
    "%Y-%m-%dT%H:%M",  # this is a custom format I added to get my form widgets working.
    "%Y-%m-%d %H:%M:%S",  # '2006-10-25 14:30:59'
    "%Y-%m-%d %H:%M:%S.%f",  # '2006-10-25 14:30:59.000200'
    "%Y-%m-%d %H:%M",  # '2006-10-25 14:30'
    "%Y-%m-%d",  # '2006-10-25'
    "%m/%d/%Y %H:%M:%S",  # '10/25/2006 14:30:59'
    "%m/%d/%Y %H:%M:%S.%f",  # '10/25/2006 14:30:59.000200'
    "%m/%d/%Y %H:%M",  # '10/25/2006 14:30'
    "%m/%d/%Y",  # '10/25/2006'
    "%m/%d/%y %H:%M:%S",  # '10/25/06 14:30:59'
    "%m/%d/%y %H:%M:%S.%f",  # '10/25/06 14:30:59.000200'
    "%m/%d/%y %H:%M",  # '10/25/06 14:30'
    "%m/%d/%y",  # '10/25/06'
]

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
# !!! Consider removing in the future.
USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/
# collect by running 'python manage.py collectstatic'
STATIC_URL = "/static/"
STATIC_ROOT = settings.django_directory / "static"

# Extra places for collectstatic to find static files.
STATICFILES_DIRS = [
    get_directory(
        settings.config_directory / "static_files"
    ),  # let's user add their own files
    settings.django_directory / "static_files",
]

# For the dynamically-created structure files, we need to include the static
# directory this to work during local testing. This is NOT allowed in a
# production server, so we don't include it when DEBUG is set to False.
# if DEBUG:
#     STATICFILES_DIRS += [STATIC_ROOT]

# This sets the django-crispy formating style
CRISPY_TEMPLATE_PACK = "bootstrap4"

# These settings help configure djangorestframework and our REST API
REST_FRAMEWORK = {
    # The REST framework has both authentication AND permission classes. Adding
    # the auth classes ensure we have user data available when rendering the
    # template. Then I allow anyone (anonymous and signed in users) to access
    # the REST APIs. Note, even though this is allow-all, we still have the
    # "RequireLoginMiddleware" applied elsewhere if the admin choses.
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.BasicAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ],
    "DEFAULT_PERMISSION_CLASSES": [
        "rest_framework.permissions.AllowAny",
    ],
    # Because we have a massive number of results for different endpoints,
    # we want to set results to be paginated by 20 results per page. This
    # way we don't have to set a page limit for every individual endpoint. Note
    # I can consider switching to LimitOffsetPagination in the future, which
    # allows the number of results per page to vary, but I don't do this
    # for now because there's no easy way to set paginator.max_limit in the
    # settings here.
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "PAGE_SIZE": 20,
    # To prevent users from querying too much and bringing down our servers,
    # we set a throttle rate on each user. Here, "anon" represents an anonymous
    # user (not signed-in) while signed-in users have access to higher download
    # rates. Note these are very restrictive because we prefer users to download
    # full databases and use Simmate locally instead.
    "DEFAULT_THROTTLE_CLASSES": [
        "rest_framework.throttling.AnonRateThrottle",
        "rest_framework.throttling.UserRateThrottle",
    ],
    "DEFAULT_THROTTLE_RATES": {"anon": "100/hour", "user": "1000/hr"},
    # We use django-filter to automatically handle filtering from a REST url
    "DEFAULT_FILTER_BACKENDS": [
        "django_filters.rest_framework.DjangoFilterBackend",
        "rest_framework.filters.OrderingFilter",
    ],
    # There are multiple ways to render the data, where we default to a nice HTML
    # view, but also support pure JSON or using Django-REST's interactive API
    "DEFAULT_RENDERER_CLASSES": [
        "rest_framework.renderers.TemplateHTMLRenderer",
        "rest_framework.renderers.BrowsableAPIRenderer",
        "rest_framework.renderers.JSONRenderer",
    ],
}

# Allows the use of iFrames from within Simmate (such as the structure-viewer)
X_FRAME_OPTIONS = "SAMEORIGIN"

# -----------------------------------------------------------------------------

# Extra settings for django-allauth

AUTHENTICATION_BACKENDS = [
    # Needed to login by username in Django admin, regardless of `allauth`
    "django.contrib.auth.backends.ModelBackend",
    # `allauth` specific authentication methods, such as login by e-mail
    "allauth.account.auth_backends.AuthenticationBackend",
]

# simple setting required by allauth. not sure what it does...
# I think this is if you are hosting several websites that all use the
# same django backend for account sign-in
SITE_ID = 1

# We start with the providers as an empty dictionary and only fill them
# if client_id and secrets are supplied in as env variables. We do this
# because we don't want broken links when users first start their server.
SOCIALACCOUNT_PROVIDERS = {}
# Other authentication plugins to consider
# "allauth.socialaccount.providers.digitalocean"
# "allauth.socialaccount.providers.orcid"

_oauth = settings.website.social_oauth

# Sign-in via Google accounts
if _oauth.google.client_id and _oauth.google.secret:
    INSTALLED_APPS.append("allauth.socialaccount.providers.google")
    SOCIALACCOUNT_PROVIDERS["google"] = {
        "APP": {
            "client_id": _oauth.google.client_id,
            "secret": _oauth.google.secret,
        },
    }

# Sign-in via Github accounts
if _oauth.github.client_id and _oauth.github.secret:
    INSTALLED_APPS.append("allauth.socialaccount.providers.github")
    SOCIALACCOUNT_PROVIDERS["github"] = {
        "APP": {
            "client_id": _oauth.github.client_id,
            "secret": _oauth.github.secret,
        },
    }

# Sign-in via Microsoft accounts
if _oauth.microsoft.client_id and _oauth.microsoft.secret:
    INSTALLED_APPS.append("allauth.socialaccount.providers.microsoft")
    SOCIALACCOUNT_PROVIDERS["microsoft"] = {
        "APP": {
            "client_id": _oauth.microsoft.client_id,
            "secret": _oauth.microsoft.secret,
            "key": "",
        },
        "TENANT": "organizations",  # limits to internal use
    }

# Initiate social login immediately -- rather than jumping to a separate
# page and then posting.
SOCIALACCOUNT_LOGIN_ON_GET = True
# SECURITY: consider removing per django-allauth's recommendation

# options for login/logoff views
LOGIN_REDIRECT_URL = "/accounts/profile/"  # this is already the default
LOGOUT_REDIRECT_URL = "/accounts/loginstatus/"

# By default, we turn off email verification. If you wish, you can switch
# this to "optional" or "mandatory"
ACCOUNT_EMAIL_VERIFICATION = os.getenv("ACCOUNT_EMAIL_VERIFICATION", "none")

# -----------------------------------------------------------------------------

# Many auth endpoints require a https instead of http response:
#   https://stackoverflow.com/questions/67726070/
# We only use this in production
if not settings.website.debug:
    ACCOUNT_DEFAULT_HTTP_PROTOCOL = "https"

if settings.website.require_login:
    MIDDLEWARE.append("simmate.website.require_login.RequireLoginMiddleware")

LOGIN_REQUIRED_URLS = (r"/(.*)$",)
LOGIN_REQUIRED_URLS_EXCEPTIONS = (
    r"/accounts(.*)$",
    r"/admin(.*)$",
    r"/static(.*)$",
    *settings.website.require_login_exceptions,
)

# -----------------------------------------------------------------------------

# BUG: To use Django ORM within IPython, Spyder, and Jupyter notebooks, which
# are examples of async consoles, I need to allow unsafe.
# Read more at...
# https://docs.djangoproject.com/en/5.0/topics/async/#async-safety
os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)


class StaticEnergy__Vasp__MatprojPbesol(StaticEnergy__Vasp__Matproj):
    """
    This task is a reimplementation of pymatgen's
    [MPStaticSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPStaticSet).

    The pymatgen library doesn't have a PBEsol static energy, so this is the
    make-shift alternative.
    """

    # tell VASP to use PBEsol functional
    _incar_updates = dict(GGA="PS")



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.matproj_hse import (
    Relaxation__Vasp__MatprojHse,
)
from simmate.apps.vasp.workflows.electronic_structure import VaspDensityOfStates


class ElectronicStructure__Vasp__MatprojDensityOfStatesHse(
    VaspDensityOfStates, Relaxation__Vasp__MatprojHse
):
    """
    This task is a reimplementation of pymatgen's
    [MPHSERelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPHSERelaxSet)
    with mode="uniform".

    Calculates the band structure using Materials Project HSE settings.
    """

    parent_workflows = ["electronic-structure.vasp.matproj-hse-full"]
    _incar_updates = dict(
        NSW=0,
        ISMEAR=0,
        SIGMA=0.05,
        ISYM=3,
        LCHARG=False,
        NELMIN=5,
    )



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.apps.bader.workflows import (
    PopulationAnalysis__Bader__Bader,
    PopulationAnalysis__Bader__CombineChgcars,
)
from simmate.apps.materials_project.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)
from simmate.engine import Workflow
from simmate.toolkit import Structure
from simmate.utilities import copy_files_from_directory


class PopulationAnalysis__VaspBader__BaderMatproj(Workflow):
    """
    Runs a static energy calculation using an extra-fine FFT grid and then
    carries out Bader analysis on the resulting charge density.
    """

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: Path = None,
        **kwargs,
    ):
        prebader_dir = directory / StaticEnergy__Vasp__PrebaderMatproj.name_full
        StaticEnergy__Vasp__PrebaderMatproj.run(
            structure=structure,
            command=command,
            source=source,
            directory=prebader_dir,
        ).result()

        # Setup chargecars for the bader analysis and wait until complete
        chgcomb_dir = directory / PopulationAnalysis__Bader__CombineChgcars.name_full
        PopulationAnalysis__Bader__CombineChgcars.run(
            directory=chgcomb_dir,
            previous_directory=prebader_dir,
        ).result()

        # And run the bader analysis on the resulting chg denisty
        bader_dir = directory / PopulationAnalysis__Bader__Bader.name_full
        PopulationAnalysis__Bader__Bader.run(
            directory=bader_dir,
            previous_directory=chgcomb_dir,
        ).result()

        # The from_vasp_directory method that loads results into the database
        # requires the following files to be in the main directory:
        #  1. the ACF.dat
        #  2. INCAR
        #  3. vasprun.xml
        #  4. POTCAR
        #  5. CHGCAR
        copy_files_from_directory(
            files_to_copy=["ACF.dat"],
            directory_new=directory,
            directory_old=bader_dir,
        )
        copy_files_from_directory(
            files_to_copy=["INCAR", "vasprun.xml", "POTCAR", "CHGCAR"],
            directory_new=directory,
            directory_old=prebader_dir,
        )
        # !!! I need a better way to access these files in the workup method
        # without copying them into the main dir...


class StaticEnergy__Vasp__PrebaderMatproj(StaticEnergy__Vasp__Matproj):
    """
    Runs a static energy calculation with a high-density FFT grid under settings
    from the Materials Project. Results can be used for Bader analysis.

    We do NOT recommend running this calculation on its own. Instead, you should
    use the full workflow, which runs this calculation AND the following bader
    analysis for you. This S3Task is only the first step of that workflow.

    See `bader.workflows.materials_project`.
    """

    # The key thing for bader analysis is that we need a very fine FFT mesh. Other
    # than that, it's the same as a static energy calculation.
    _incar_updates = dict(
        NGXF__density_a=20,
        NGYF__density_b=20,
        NGZF__density_c=20,
        LAECHG=True,  # write core charge density to AECCAR0 and valence to AECCAR2
    )



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .all_paths import NebAllPathsWorkflow
from .from_endpoints import NebFromEndpointWorkflow
from .from_images import VaspNebFromImagesWorkflow
from .single_path import SinglePathWorkflow



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .figure import Figure

from .matplotlib import MatplotlibFigure
from .plotly import PlotlyFigure



# -*- coding: utf-8 -*-

import logging
import math
import traceback
from pathlib import Path

import numpy
import pandas
import plotly.express as plotly_express
import plotly.graph_objects as plotly_go
from rich.progress import track

from simmate.apps.evolution import selectors as selector_module
from simmate.apps.evolution.models import SteadystateSource
from simmate.configuration.dask import get_dask_client
from simmate.database.base_data_types import Calculation, table_column
from simmate.engine.execution import WorkItem
from simmate.toolkit import Composition, Structure
from simmate.toolkit.validators import fingerprint as validator_module
from simmate.utilities import get_directory
from simmate.visualization.plotting import PlotlyFigure


class FixedCompositionSearch(Calculation):
    """
    This database table holds all of the information related to an evolutionary
    search and also has convient methods to analyze the data + write output files.
    """

    class Meta:
        app_label = "workflows"

    # !!! consider making a composition-based mixin
    composition = table_column.CharField(max_length=50, null=True, blank=True)

    # Import path for the workflow and the database table of results
    subworkflow_name = table_column.CharField(max_length=200, null=True, blank=True)
    subworkflow_kwargs = table_column.JSONField(default=dict, null=True, blank=True)
    fitness_field = table_column.CharField(max_length=200, null=True, blank=True)

    # Other settings for the search
    min_structures_exact = table_column.IntegerField(null=True, blank=True)
    max_structures = table_column.IntegerField(null=True, blank=True)
    best_survival_cutoff = table_column.IntegerField(null=True, blank=True)
    nfirst_generation = table_column.IntegerField(null=True, blank=True)
    nsteadystate = table_column.IntegerField(null=True, blank=True)
    convergence_cutoff = table_column.FloatField(null=True, blank=True)

    # Key classes to use during the search
    selector_name = table_column.CharField(max_length=200, null=True, blank=True)
    selector_kwargs = table_column.JSONField(default=dict, null=True, blank=True)
    validator_name = table_column.CharField(max_length=200, null=True, blank=True)
    validator_kwargs = table_column.JSONField(default=dict, null=True, blank=True)
    singleshot_sources = table_column.JSONField(default=list, null=True, blank=True)
    # stop_condition_name ---> assumed for now
    # information about the steadystate_sources are stored within
    # the SteadstateSource datatable

    # the time to sleep between file writing and steady-state checks.
    sleep_step = table_column.FloatField(null=True, blank=True)

    # Loading all unique structures is an expensive operation so its only
    # done on a cycle. This stores the list of unique ids.
    unique_individuals_ids = table_column.JSONField(
        default=list,
        null=True,
        blank=True,
    )

    # This is an optional an input for an expected structure in order to allow
    # creation of plots that show convergence vs. the expected. This is very
    # useful benchmarking and when the user has a target structure that they
    # hope to see. For now, this MUST a dictionary input (either pymatgen.as_dict()
    # that points to another table entry because we don't want to use the
    # Structure mix-in on this table (too many unecessary columns)
    expected_structure = table_column.JSONField(default=dict)

    # TODO:
    #   parent_variable_nsite_searches
    #   parent_binary_searches
    #   parent_ternary_searches

    # -------------------------------------------------------------------------
    # Setup of the database tables at the beginning of a new search
    # -------------------------------------------------------------------------

    @classmethod
    def from_toolkit(
        cls,
        steadystate_sources: dict = None,
        as_dict: bool = False,
        **kwargs,
    ):
        # Initialize the steady state sources by saving their config information
        # to the database.
        if steadystate_sources and not as_dict:
            search = cls(**kwargs)
            search.save()  # we must save up front because of relations
            search._init_steadystate_sources_to_db(steadystate_sources)
            return search

        elif not steadystate_sources:
            return kwargs if as_dict else cls(**kwargs)

        if steadystate_sources and as_dict:
            raise Exception(
                "steadystate_sources cannot be set in an as_dict mannor because"
                "it points to a related database table"
            )

    # -------------------------------------------------------------------------
    # Highest-level methods that run the overall search
    # -------------------------------------------------------------------------

    def check_stop_condition(self):
        # First, see if we have at least our minimum limit for *exact* structures.
        # "Exact" refers to the nsites of the structures. We want to ensure at
        # least N structures with the input/expected number of sites have been
        # calculated.
        # {f"{self.fitness_field}__isnull"=False} # when I allow other fitness fxns
        count_exact = self.individuals_datatable.objects.filter(
            formula_full=self.composition,
            workflow_name=self.subworkflow_name,
            energy_per_atom__isnull=False,
        ).count()
        if count_exact < self.min_structures_exact:
            return False

        # Next, see if we've hit our maximum limit for structures.
        # Note: because we are only looking at the results table, this is really
        # only counting the number of successfully calculated individuals.
        # Nothing is done to stop those that are still running or to count
        # structures that failed to be calculated
        # {f"{self.fitness_field}__isnull"=False} # when I allow other fitness fxns
        if self.individuals_completed.count() > self.max_structures:
            logging.info(
                "Maximum number of completed calculations hit "
                f"(n={self.max_structures})."
            )
            return True

        # The next stop condition is based on how long we've have the same
        # "best" individual. If the number of new individuals calculated (without
        # any becoming the new "best") is greater than best_survival_cutoff, then
        # we can stop the search.
        # grab the best individual for reference
        best = self.best_individual

        # We need this if-statement in case no structures have completed yet.
        if not best:
            return False

        # count the number of new individuals added AFTER the best one. If it is
        # more than best_survival_cutoff, we stop the search.
        # Note, we look at all structures that have an energy_per_atom greater
        # than 1meV/atom higher than the best structure. The +1meV ensures
        # we aren't prolonging the calculation for insignificant changes in
        # the best structure. Looking at the energy also ensures that we are
        # only counting completed calculations.
        # BUG: this filter needs to be updated to fitness_value and not
        # assume we are using energy_per_atom
        num_new_structures_since_best = self.individuals.filter(
            energy_per_atom__gt=best.energy_per_atom + self.convergence_cutoff,
            finished_at__gte=best.finished_at,
        ).count()
        if num_new_structures_since_best > self.best_survival_cutoff:
            logging.info(
                "Best individual has not changed after "
                f"{self.best_survival_cutoff} new individuals added."
            )
            return True
        # If we reached this point, then we haven't hit a stop condition yet!
        return False

    def _check_singleshot_sources(self, directory: Path):
        # local imports to prevent circuluar import issues
        from simmate.apps.evolution.singleshot_sources.known import get_known_structures
        from simmate.apps.evolution.singleshot_sources.prototypes import (
            get_structures_from_prototypes,
        )
        from simmate.apps.evolution.singleshot_sources.substitution import (
            get_structures_from_substitution_of_known,
        )
        from simmate.apps.evolution.workflows.utilities import (
            write_and_submit_structures,
        )

        composition = Composition(self.composition)

        if "third_parties" in self.singleshot_sources:
            structures_known = get_known_structures(
                composition,
                allow_multiples=False,
                remove_matching=True,
            )
            logging.info(
                f"Generated {len(structures_known)} structures from third-party databases"
            )
            write_and_submit_structures(
                structures=structures_known,
                foldername=directory / "from_third_parties",
                workflow=self.subworkflow,
                workflow_kwargs=self.subworkflow_kwargs,
            )

        if "third_party_substituition" in self.singleshot_sources:
            structures_sub = get_structures_from_substitution_of_known(
                composition,
                allow_multiples=False,
            )
            logging.info(
                f"Generated {len(structures_sub)} structures from substitutions"
            )
            write_and_submit_structures(
                structures=structures_sub,
                foldername=directory / "from_third_party_substituition",
                workflow=self.subworkflow,
                workflow_kwargs=self.subworkflow_kwargs,
            )

        if "prototypes" in self.singleshot_sources:
            structures_prototype = get_structures_from_prototypes(
                composition,
                max_sites=int(composition.num_atoms),
                remove_matching=True,
            )
            logging.info(
                f"Generated {len(structures_prototype)} structures from prototypes"
            )
            write_and_submit_structures(
                structures=structures_prototype,
                foldername=directory / "from_prototypes",
                workflow=self.subworkflow,
                workflow_kwargs=self.subworkflow_kwargs,
            )

    def _init_steadystate_sources_to_db(self, steadystate_sources):
        composition = Composition(self.composition)

        # Initialize the steady-state sources, which are given as a list of
        # (proportion, class/class_str, kwargs) for each. As we go through
        # these, we also collect the proportion list for them.
        steadystate_sources_cleaned = []
        steadystate_source_proportions = []
        for source_name, proportion in steadystate_sources.items():
            # There are certain transformation sources that don't work for
            # single-element structures, so we check for this here and
            # remove them.
            if len(composition.elements) == 1 and source_name in [
                "from_ase.AtomicPermutation"
            ]:
                logging.warning(
                    f"{source_name} is not possible with single-element structures."
                    " This is being removed from your steadystate_sources."
                )
                continue  # skips to next source

            # Store proportion value and name of source. We do NOT initialize
            # the source because it will be called elsewhere (within the submitted
            # workflow and therefore a separate thread.
            steadystate_sources_cleaned.append(source_name)
            steadystate_source_proportions.append(proportion)

        # Make sure the proportions sum to 1, otherwise scale them. We then convert
        # these to steady-state integers (and round to the nearest integer)
        sum_proportions = sum(steadystate_source_proportions)
        if not math.isclose(sum_proportions, 1, rel_tol=1e-6):
            logging.warning(
                "fractions for steady-state sources do not add to 1. "
                "We have scaled all sources to equal one to fix this."
            )
            steadystate_source_proportions = [
                p / sum_proportions for p in steadystate_source_proportions
            ]
        # While these are percent values, we want specific counts. We convert to
        # those here.
        steadystate_source_counts = [
            int(p * self.nsteadystate) for p in steadystate_source_proportions
        ]

        # Throughout our search, we want to keep track of which workflows we've
        # submitted for each source as well as how long we were holding a
        # steady-state of submissions. We therefore keep a log of Sources in
        # our database.
        steadystate_sources_db = []
        for source_name, ntarget_jobs in zip(
            steadystate_sources_cleaned, steadystate_source_counts
        ):
            # before saving, we want to record whether we have a creator or transformation
            # !!! Is there a way to dynamically determine this from a string?
            # Or maybe initialize them to provide they work before we register?
            # Third option is to specify with the input.
            known_creators = [
                "RandomSymStructure",
                "PyXtalStructure",
            ]
            is_creator = True if source_name in known_creators else False
            known_transformations = [
                "ExtremeSymmetry",
                "from_ase.Heredity",
                "from_ase.SoftMutation",
                "from_ase.MirrorMutation",
                "from_ase.LatticeStrain",
                "from_ase.RotationalMutation",
                "from_ase.AtomicPermutation",
                "from_ase.CoordinatePerturbation",
            ]
            is_transformation = True if source_name in known_transformations else False
            if not is_creator and not is_transformation:
                raise Exception("Unknown source being used.")
                # BUG: I should really allow any subclass of Creator/Transformation

            source_db = SteadystateSource(
                name=source_name,
                is_creator=is_creator,
                is_transformation=is_transformation,
                nsteadystate_target=ntarget_jobs,
                search=self,
            )
            source_db.save()
            steadystate_sources_db.append(source_db)
        # !!! What if this search has been ran before or a matching search is
        # being ran elsewhere? Do we still want new entries for each?

        return steadystate_sources_db

    def _check_steadystate_workflows(self):
        # local import to prevent circular import issues
        from simmate.apps.evolution.workflows.new_individual import (
            StructurePrediction__Toolkit__NewIndividual,
        )

        # transformations from a database table require that we have
        # completed structures in the database. We want to wait until there's
        # a set amount before we start mutating the best. We check that here.
        if self.individuals_completed.count() < self.nfirst_generation:
            logging.info(
                "Search hasn't finished nfirst_generation yet "
                f"({self.nfirst_generation} individuals). "
                "Skipping transformations."
            )
            ready_for_transformations = False
        else:
            ready_for_transformations = True

        # we iterate through each steady-state source and check to see how many
        # jobs are still running for it. If it's less than the target steady-state,
        # then we need to submit more!
        steadystate_sources_db = self.steadystate_sources.all()
        for source_db in steadystate_sources_db:
            # skip if we have a transformation but aren't ready for it yet
            if source_db.is_transformation and not ready_for_transformations:
                continue

            # This loop says for the number of steady state runs we are short,
            # create that many new individuals! max(x,0) ensure we don't get a
            # negative value. A value of 0 means we are at steady-state and can
            # just skip this loop.
            nflows_to_submit = max(
                int(source_db.nsteadystate_target - source_db.nflow_runs), 0
            )

            if nflows_to_submit > 0:
                logging.info(
                    f"Submitting {nflows_to_submit} new individuals for "
                    f"'{source_db.name}'"
                )

            # disable the logs while we submit
            logger = logging.getLogger()
            logger.disabled = True

            for n in range(nflows_to_submit):
                # submit the workflow for the new individual. Note, the structure
                # won't be evuluated until the job actually starts. This allows
                # our validator to have the most current information available
                # when starting the structure creation
                state = StructurePrediction__Toolkit__NewIndividual.run_cloud(
                    search_id=self.id,
                    steadystate_source_id=source_db.id,
                    # NOTE: we *only* pass tags because that is all that is
                    # relevent for submission. All other kwargs are passed
                    # within the new-individual workflow to the expected subflow
                    tags=self.subworkflow_kwargs.get("tags", []),
                )

                # Attached the id to our source so we know how many
                # associated jobs are running.
                # NOTE: this is the WorkItem id and NOT the run_id!!!
                source_db.workitem_ids.append(state.pk)
                source_db.save()

            # reactivate logging
            logger.disabled = False

    # -------------------------------------------------------------------------
    # Core methods that help grab key information about the search
    # -------------------------------------------------------------------------

    @property
    def subworkflow(self):
        # local import to prevent circular import issues
        from simmate.workflows.utilities import get_workflow

        # Initialize the workflow if a string was given.
        # Otherwise we should already have a workflow class.
        if self.subworkflow_name == "relaxation.vasp.staged":
            workflow = get_workflow(self.subworkflow_name)
        else:
            raise Exception(
                "Only `relaxation.vasp.staged` is supported in early testing"
            )

        # BUG: I'll need to rewrite this in the future bc I don't really account
        # for other workflows yet. It would make sense that our workflow changes
        # as the search progresses (e.g. we incorporate DeePMD relaxation once
        # ready)

        return workflow

    @property
    def individuals_datatable(self):
        # NOTE: this table just gives the class back and doesn't filter down
        # to the relevent individuals for this search. For that, use the
        # "individuals" property
        # we assume the table is registered in the local_calcs app
        return self.subworkflow.database_table

    @property
    def individuals(self):
        # note we don't call "all()" on this queryset yet becuase this property
        # it often used as a "base" queryset (and additional filters are added)
        composition = Composition(self.composition)
        return self.individuals_datatable.objects.filter(
            # You'd expect this filter to be...
            #   formula_full=self.composition
            # However, this misses structures that are reduced to a smaller
            # unitcells during relaxation. Therefore, by default, we need to
            # include all individuals that have fewer nsites. This is
            # often what a user wants anyways during searches, so it works out.
            formula_reduced=composition.reduced_formula,
            nsites__lte=composition.num_atoms,
            workflow_name=self.subworkflow_name,
        )

    @property
    def individuals_completed(self):
        # If there is an energy_per_atom, we can treat the calculation as completed
        return self.individuals.filter(energy_per_atom__isnull=False)
        # OPTIMIZE: would it be better to check energy_per_atom or structure_final?
        # Ideally, I could make a relation to the prefect flow run table but this
        # would require a large amount of work to implement.

    @property
    def individuals_incomplete(self):
        # If there is an energy_per_atom, we can treat the calculation as completed
        return self.individuals.filter(energy_per_atom__isnull=True)

    @property
    def best_individual(self):
        return self.individuals_completed.order_by(self.fitness_field).first()

    def get_nbest_indiviudals(self, nbest: int):
        return self.individuals_completed.order_by(self.fitness_field)[:nbest]

    def get_unique_individuals(
        self,
        use_cache: bool = False,
        as_queryset: bool = False,
    ):
        # get the most-up-to-date results but is slow
        if not use_cache:
            unique = self.validator.get_unique_from_pool()
            # This is an expensive method as the calculation scales, so make sure
            # we cache thes values
            self.unique_individuals_ids = [s.database_object.id for s in unique]
            self.save()

        # OPTIMIZE: this converts back to a queryset object but involves
        # another database query.
        if as_queryset or use_cache:
            unique = (
                self.individuals_completed.filter(id__in=self.unique_individuals_ids)
                .order_by(self.fitness_field)
                .all()
            )

        if not as_queryset and not isinstance(unique, list):
            unique = unique.to_toolkit()

        return unique

    def get_best_individual_history(self):
        """
        Goes through all structures in order that they were created and creates
        a history of which structure was best at any given time.
        """

        individuals = (
            self.individuals_completed.order_by("finished_at")
            .only("id", self.fitness_field)
            .all()
        )

        # Keep a log of the best structures. The first structure to finish is
        # by default the best at that time.
        best_history = [individuals[0].id]
        best_value = getattr(individuals[0], self.fitness_field)
        for individual in individuals:
            potential_new_value = getattr(individual, self.fitness_field)

            # BUG: I assume lower is better, but this may change
            if potential_new_value < best_value:
                best_history.append(individual.id)
                best_value = potential_new_value

        return best_history

    def write_output_summary(self, directory: Path):
        logging.info(f"Writing search summary to {directory}")

        super().write_output_summary(directory=directory)

        # If the output fails to write, we have a non-critical issue that
        # doesn't affect the search. We therefore don't want to raise an
        # error here -- but instead warn the user and then continue the search
        try:
            # calls all the key methods defined below
            best_cifs_directory = get_directory(directory / "best_structures")
            self.write_best_structures(100, best_cifs_directory)
            unique_cifs_directory = get_directory(directory / "best_structures_unique")
            self.write_unique_structures(unique_cifs_directory)

            self.write_individuals_completed(directory=directory)
            self.write_individuals_completed_full(directory=directory)
            self.write_best_individuals_history(directory=directory)
            self.write_individuals_incomplete(directory=directory)
            self.write_fitness_convergence_plot(directory=directory)
            self.write_fitness_distribution_plot(directory=directory)
            self.write_subworkflow_times_plot(directory=directory)

            # BUG: This is only for "relaxation.vasp.staged", which the assumed
            # workflow for now.
            self.write_staged_series_convergence_plot(directory=directory)
            self.write_staged_series_histogram_plot(directory=directory)
            self.write_staged_series_times_plot(directory=directory)

            logging.info("Done writing summary.")

        except Exception:
            logging.warning(
                "Failed to write the output summary. This issue will be silenced "
                "to avoid stopping the search. But please report the following "
                "error to our github: https://github.com/jacksund/simmate/issues/"
            )

            # prints the most recent exception traceback
            traceback.print_exc()

    # -------------------------------------------------------------------------
    # Methods for deserializing objects. Consider moving to the toolkit methods
    # -------------------------------------------------------------------------

    @property
    def selector(self):
        # !!! This is largley a copy/paste of the validator method. Consider
        # making a utility to load classes by name

        if hasattr(selector_module, self.selector_name):
            selector_class = getattr(selector_module, self.selector_name)
            selector = selector_class(**self.selector_kwargs)
        else:
            raise Exception("Unknown selector name provided")

        return selector

    @property
    def validator(self):
        # Initialize the fingerprint database
        # For this we need to grab all previously calculated structures of this
        # compositon too pass in too.

        validator_class = getattr(validator_module, self.validator_name)

        logging.info("Generating fingerprints for past structures...")
        fingerprint_validator = validator_class(
            composition=Composition(self.composition),
            structure_pool=self.individuals_completed.order_by(self.fitness_field),
            **self.validator_kwargs,
        )
        logging.info("Done generating fingerprints.")

        # for now I only include final structures that have been calculated
        # OPTIMIZE: should we only do structures that were successfully calculated?
        # If not, there's a chance a structure fails because of something like a
        # crashed slurm job, but it's never submitted again...
        # Or should we only do final structures? Or should I include input
        # structures and even all ionic steps as well...?

        return fingerprint_validator

    # -------------------------------------------------------------------------
    # Writing CSVs summaries and CIFs of best structures
    # -------------------------------------------------------------------------

    def write_best_structures(self, nbest: int, directory: Path):
        best = self.get_nbest_indiviudals(nbest)
        structures = best.only("structure", "id").to_toolkit()
        self._write_structures(structures, directory)

    def write_unique_structures(self, directory: Path):
        structures = self.get_unique_individuals(use_cache=False)
        self._write_structures(structures, directory)

    # TODO: consider making a utility elsewhere
    def _write_structures(self, structures: list[Structure], directory: Path):
        # if the directory is filled, we need to delete all the files
        # before writing the new ones.
        for file in directory.iterdir():
            try:
                file.unlink()
            except OSError:
                logging.warning("Unable to delete a CIF file: {file}")
                logging.warning(
                    "Updating this directory involves deleting "
                    "and re-writing all CIF files each cycle. If you have a file "
                    "open while this step occurs, then you'll see this warning."
                    "Close your file for this to go away."
                )

        for rank, structure in enumerate(structures):
            rank_cleaned = str(rank).zfill(2)  # converts 1 to 01
            structure_filename = (
                directory
                / f"rank-{str(rank_cleaned)}__id-{structure.database_object.id}.cif"
            )
            structure.to(filename=str(structure_filename), fmt="cif")

    def write_individuals_completed_full(self, directory: Path):
        columns = self.individuals_datatable.get_column_names()
        columns.remove("structure")
        df = self.individuals_completed.defer("structure").to_dataframe(columns)
        csv_filename = directory / "individuals_completed__ALLDATA.csv"
        df.to_csv(csv_filename)

    def write_individuals_completed(self, directory: Path):
        columns = [
            "id",
            "energy_per_atom",
            "finished_at",
            "source",
            "spacegroup__number",
        ]
        df = (
            self.individuals_completed.order_by(self.fitness_field)
            .only(*columns)
            .to_dataframe(columns)
        )
        # label the index column
        df.index.name = "rank"

        # make the timestamps easier to read
        def format_date(date):
            return date.strftime("%Y-%m-%d %H:%M:%S")

        df["finished_at"] = df.finished_at.apply(format_date)

        def format_parents(source):
            return source.get("parent_ids", None) if source else None

        df["parent_ids"] = df.source.apply(format_parents)

        def format_source(source):
            return (
                None
                if not source
                else source.get("creator", None) or source.get("transformation", None)
            )

        df["source"] = df.source.apply(format_source)

        # shorten the column name for easier reading
        df.rename(columns={"spacegroup__number": "spacegroup"}, inplace=True)

        md_filename = directory / "individuals_completed.md"
        df.to_markdown(md_filename)

    def write_best_individuals_history(self, directory: Path):
        columns = ["id", "energy_per_atom", "finished_at"]
        best_history = self.get_best_individual_history()
        df = (
            self.individuals.filter(id__in=best_history)
            .order_by("-finished_at")
            .only(*columns)
            .to_dataframe(columns)
        )

        # make the timestamps easier to read
        def format_date(date):
            return date.strftime("%Y-%m-%d %H:%M:%S")

        df["finished_at"] = df.finished_at.apply(format_date)
        md_filename = directory / "history_of_the_best_individuals.md"
        df.to_markdown(md_filename)

    def write_individuals_incomplete(self, directory: Path):
        structure_sources = self.steadystate_sources.all()
        sources = []
        workitem_ids = []
        statuses = []
        created_at = []
        for source in structure_sources:
            workitem_ids += source.workitem_ids
            # as an extra, keep a list of the source names
            sources += [source.name] * len(source.workitem_ids)

            for workitem_id in source.workitem_ids:
                work_item = WorkItem.objects.get(id=workitem_id)
                statuses.append(work_item.get_status_display())
                created_at.append(work_item.created_at.strftime("%Y-%m-%d %H:%M:%S"))

        df = pandas.DataFrame(
            {
                "workitem_id": workitem_ids,
                "source": sources,
                "status": statuses,
                "created_at": created_at,
            }
        )

        md_filename = directory / "individuals_still_running.md"
        df.to_markdown(md_filename)


class FitnessConvergence(PlotlyFigure):
    def get_plot(search: FixedCompositionSearch):
        # Grab the calculation's structure and convert it to a dataframe
        columns = ["finished_at", search.fitness_field]
        structures_dataframe = search.individuals_completed.only(*columns).to_dataframe(
            columns
        )

        # There's only one plot here, no subplot. So we make the scatter
        # object and just pass it directly to a Figure object
        figure = plotly_express.scatter(
            x=structures_dataframe["finished_at"],
            y=structures_dataframe[search.fitness_field],
            marginal_x="histogram",
            marginal_y="histogram",
        )

        figure.update_layout(
            xaxis_title="Date Completed",
            yaxis_title=(
                "Energy (eV/atom)"
                if search.fitness_field == "energy_per_atom"
                else search.fitness_field
            ),
        )

        return figure


class Correctness(PlotlyFigure):
    def get_plot(
        search: FixedCompositionSearch,
        structure_known: Structure,
    ):
        # load the featurizer from the search object
        featurizer = search.validator.featurizer

        # Grab the calculation's structure and convert it to a dataframe
        structures_dataframe = search.individuals_completed.to_dataframe()

        # because we are using the database model, we first want to convert to
        # pymatgen structures objects and add a column to the dataframe for these
        #
        #   structures_dataframe["structure"] = [
        #       structure.to_toolkit() for structure in ionic_step_structures
        #   ]
        #
        # BUG: the read_frame query creates a new query, so it may be a different
        # length from ionic_step_structures. For this reason, we can't iterate
        # through the queryset like in the commented out code above. Instead,
        # we need to iterate through the dataframe rows.
        # See https://github.com/chrisdev/django-pandas/issues/138 for issue

        structures_dataframe["structure"] = [
            Structure.from_str(s.structure, fmt="POSCAR")
            for _, s in structures_dataframe.iterrows()
        ]

        # generating fingerprints is slow so we use dask to parallelize
        client = get_dask_client()
        logging.info("Submitting to jobs dask...")
        futures = [
            client.submit(featurizer.featurize, s.structure, pure=False)
            for _, s in track(list(structures_dataframe.iterrows()))
        ]
        logging.info("Waiting for dask jobs to finish...")
        structures_dataframe["fingerprint"] = [numpy.array(f.result()) for f in futures]
        logging.info("Done.")

        fingerprint_known = numpy.array(featurizer.featurize(structure_known))

        distances = []
        smallest_distance_id = None
        smallest_distance = 999
        for _, s in structures_dataframe.iterrows():
            distance = numpy.linalg.norm(fingerprint_known - s.fingerprint)
            distances.append(distance)

            # import scipy
            # scipy.spatial.distance.cosine(fingerprint_known, s.fingerprint)
            # BUG: I assume distance method. I need to check the validator in
            # case the method prefers something like cosine distance.

            if distance < smallest_distance:
                smallest_distance = distance
                smallest_distance_id = s.id
        logging.info(
            f"The most similar structure is id={smallest_distance_id} "
            f"with distance {smallest_distance:.3f}"
        )
        structures_dataframe["fingerprint_distance"] = distances

        # There's only one plot here, no subplot. So we make the scatter
        # object and just pass it directly to a Figure object
        scatter = plotly_go.Scatter(
            x=structures_dataframe["finished_at"],
            y=structures_dataframe["fingerprint_distance"],
            mode="markers",
            marker_color=structures_dataframe["energy_per_atom"],
        )
        figure = plotly_go.Figure(data=scatter)

        figure.update_layout(
            xaxis_title="Date Completed",
            yaxis_title="Distance from Known Structure",
        )

        return figure


class SubworkflowTimes(PlotlyFigure):
    def get_plot(search: FixedCompositionSearch):
        # Grab the calculation's structure and convert it to a dataframe
        data = search.individuals_completed.values_list(
            "total_time", "queue_time"
        ).all()

        # time is stored in seconds and we convert to minutes
        total_times = [e[0] / 60 for e in data]
        # queue_times = [e[1] / 60 for e in data]

        figure = plotly_go.Figure()
        hist_1 = plotly_go.Histogram(x=total_times)  # , name="Total run time (min)"
        # hist_2 = plotly_go.Histogram(x=queue_times, name="Total queue time (min)")
        figure.add_trace(hist_1)
        # figure.add_trace(hist_2)
        figure.update_layout(
            xaxis_title="Total calculation time (min)",
            yaxis_title="Individuals (#)",
            barmode="overlay",
        )
        # figure.update_traces(opacity=0.75)
        return figure


class FitnessDistribution(PlotlyFigure):
    def get_plot(search: FixedCompositionSearch):
        # Grab the calculation's structure and convert it to a dataframe
        structures_dataframe = search.individuals_completed.only(
            search.fitness_field
        ).to_dataframe(search.fitness_field)

        # There's only one plot here, no subplot. So we make the scatter
        # object and just pass it directly to a Figure object
        histogram = plotly_go.Histogram(
            x=structures_dataframe[search.fitness_field],
        )
        figure = plotly_go.Figure(data=histogram)

        figure.update_layout(
            xaxis_title=(
                "Energy (eV/atom)"
                if search.fitness_field == "energy_per_atom"
                else search.fitness_field
            ),
            yaxis_title="Individuals (#)",
        )

        return figure


class StagedSeriesConvergence(PlotlyFigure):
    def get_plot(search: FixedCompositionSearch):
        composition = Composition(search.composition)
        plot = search.subworkflow.get_staged_series_convergence_plot(
            # See `individuals` method for why we use these filters
            formula_reduced=composition.reduced_formula,
            nsites__lte=composition.num_atoms,
            energy_per_atom__isnull=False,
        )
        return plot


class StagedSeriesHistogram(PlotlyFigure):
    def get_plot(search: FixedCompositionSearch):
        composition = Composition(search.composition)
        plot = search.subworkflow.get_staged_series_histogram_plot(
            # See `individuals` method for why we use these filters
            formula_reduced=composition.reduced_formula,
            nsites__lte=composition.num_atoms,
            energy_per_atom__isnull=False,
        )
        return plot


class StagedSeriesTimes(PlotlyFigure):
    def get_plot(search: FixedCompositionSearch):
        composition = Composition(search.composition)
        plot = search.subworkflow.get_staged_series_times_plot(
            # See `individuals` method for why we use these filters
            formula_reduced=composition.reduced_formula,
            nsites__lte=composition.num_atoms,
            energy_per_atom__isnull=False,
        )
        return plot


# register all plotting methods to the database table
for _plot in [
    FitnessConvergence,
    Correctness,
    FitnessDistribution,
    SubworkflowTimes,
    StagedSeriesConvergence,
    StagedSeriesHistogram,
    StagedSeriesTimes,
]:
    _plot.register_to_class(FixedCompositionSearch)



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import (
    BandStructureCalc,
    DensityofStatesCalc,
    DiffusionAnalysis,
    Dynamics,
    DynamicsIonicStep,
    IonicStep,
    MigrationHop,
    MigrationImage,
    PopulationAnalysis,
    Relaxation,
    StaticEnergy,
)



# -*- coding: utf-8 -*-

import logging
import shutil
from pathlib import Path

from simmate.apps.evolution.models.fixed_composition import FixedCompositionSearch
from simmate.engine import Workflow


class StructurePrediction__Toolkit__NewIndividual(Workflow):
    """
    Generates a new individual for an evolutionary search algorithm.

    Note, this workflow should not be called directly, but instead used within
    higher level workflows (such as `fixed-composition`).

    Users will rarely (if ever) need to call this workflow
    """

    description_doc_short = "a single structure submission for a search"

    has_prerequisite = True
    use_database = False

    @staticmethod
    def run_config(
        search_id: int,
        steadystate_source_id: int,
        directory: Path,
        **kwargs,
    ):
        search_db = FixedCompositionSearch.objects.get(id=search_id)
        source_db = search_db.steadystate_sources.get(id=steadystate_source_id)
        validator = search_db.validator

        # Check the stop condition of the search and see if this new individual
        # is even needed. This will catch when a search ends while a new
        # individual was waiting in the queue
        if search_db.check_stop_condition():
            logging.info(
                "The search ended while this individual was in the queue. "
                "Canceling new individual."
            )
            shutil.rmtree(directory)
            return

        if source_db.is_transformation:
            transformer = source_db.to_toolkit()

            # Check the cache first to save time (even though it might be outdated)
            # and then actually calculate all unique as a backup
            # We only activate the caching once we have >200 unique structures,
            # as that is when it starts to cut into CPU time.
            use_cache = bool(len(search_db.unique_individuals_ids) >= 400)
            unique_queryset = search_db.get_unique_individuals(
                use_cache=use_cache,
                as_queryset=True,
            )

            output = transformer.apply_from_database_and_selector(
                selector=search_db.selector,
                datatable=unique_queryset,
                select_kwargs=dict(
                    fitness_column=search_db.fitness_field,
                    # query_limit=200,  # OPTIMIZE: Smarter way to do this...?
                ),
                validators=[validator],
            )

            # if the source failed to create a structure, then we want to remove
            # it to prevent repeated issues.
            if output == False:
                # TODO: consider more advanced logic for changing the steady
                # state values of each source -- rather than just disabling
                # them here.
                logging.warning(
                    "Failed to create new individual with steady-state "
                    f"source {source_db.name}. Removing steady-state."
                )
                source_db.nsteadystate_target = 0
                source_db.save()
                shutil.rmtree(directory)
                return

            # otherwise we have a successful output that we can use
            parent_ids, new_structure = output
            source = {
                "transformation": source_db.name,
                "parent_ids": parent_ids,
            }

        elif source_db.is_creator:
            creator = source_db.to_toolkit()

            new_structure = creator.create_structure_with_validation(
                validators=[validator],
            )
            source = {
                "creator": source_db.name,
            }

        # if structure creation was successful, run the workflow for it
        if new_structure:
            state = search_db.subworkflow.run(
                structure=new_structure,
                source=source,
                directory=directory,  # BUG: consider making subfolder
                **search_db.subworkflow_kwargs,
            )
            result = state.result()
            # NOTE: we tell the workflow to use the same directory. There is
            # good chance the user indicates that they want to compress the
            # folder to.

            # check the final structure with our validator again. This populates
            # the fingerprint database (if one is being used).
            validator.check_structure(result.to_toolkit())
            # BUG: I think there is a race condition here... Other NewIndividual
            # workflows may try to populate the fingerprint at the START of
            # a run while this one must do it after .result() is called.

        # TODO: when I allow a series of subworkflows, I can do validation checks
        # between each run.
        # if a validator was given, we want to check the current structure
        # and see if it passes our test. This is typically only done in
        # expensive analysis -- like evolutionary searches
        # current_structure = result.to_toolkit()
        # if validator and not validator.check_structure(current_structure):
        #     # if it fails the check, we want to stop the series of calculations
        #     # and just exit the workflow run. We can, however, update the
        #     # database entry with the final structure.
        #     logging.info(
        #         "Did not pass validation checkpoint. Stopping workflow series."
        #     )
        #     return {"structure": current_structure}



# -*- coding: utf-8 -*-

import pandas

from simmate.apps.evolution.selectors import Selector


class TournamentSelection(Selector):
    """
    Tournament selection selects N parents randomly and then choses the single
    best of these those randomly selected. If multiple individuals are requested,
    a separate tournament is used to select each.
    """

    @staticmethod
    def select(
        nselect: int,
        individuals: pandas.DataFrame,
        fitness_column: str,
        # The percent of individuals participating in each tournament
        tournament_size: float = 0.20,
        tournament_min: int = 3,
    ) -> pandas.DataFrame:
        # We want the tournament size to be based on the total number of
        # inidividuals available.
        ntournament = int(len(individuals) * tournament_size)
        if ntournament < tournament_min:
            ntournament = tournament_min

        # Run the series of tournaments and select the winner from each
        tournament_winners = []
        for n in range(nselect):
            df_tourn = individuals.sample(ntournament, replace=False)
            winner = df_tourn.nsmallest(1, fitness_column).iloc[0]
            tournament_winners.append(winner)

        # take the winning series and convert back to a pandas dataframe
        df_parents = pandas.DataFrame(tournament_winners)
        df_parents = df_parents.reset_index(drop=True)  # in case of duplicates

        return df_parents



# -*- coding: utf-8 -*-

from simmate.toolkit.validators.fingerprint.base import FingerprintValidator

from .featurizers import CrystalNNFingerprint as cnnf
from .featurizers import SiteStatsFingerprint as ssf


class CrystalNNFingerprint(FingerprintValidator):
    @staticmethod
    def get_featurizer(
        stat_options: list[str] = ["mean", "std_dev", "minimum", "maximum"],
        **crystalnn_options,
    ):
        # make the matminer featurizer object
        # if it isnt set, we will use the default, which is set to what the Materials project uses
        if crystalnn_options:
            sitefingerprint_method = cnnf(**crystalnn_options)
        else:
            # see https://materialsproject.org/docs/structuresimilarity
            sitefingerprint_method = cnnf.from_preset(
                "ops", distance_cutoffs=None, x_diff_weight=0
            )
        # now that we made the sitefingerprint_method, we can input it into the
        # structurefingerprint_method which finishes up the featurizer
        featurizer = ssf(sitefingerprint_method, stats=stat_options)
        return featurizer



# -*- coding: utf-8 -*-

from simmate.apps.vasp.inputs import Poscar


def test_poscar(tmp_path, structure):
    filename = tmp_path / "POSCAR"
    Poscar.to_file(structure, filename)
    structure_new = Poscar.from_file(filename)
    assert structure == structure_new



# -*- coding: utf-8 -*-

"""
ASE Transformations
--------------------

All transformations in this module are originally from the 
[ASE](https://wiki.fysik.dtu.dk/ase/) package. These classes simply wrap ASE's
original code so that their use matches the other Simmate transformations.
"""

from .atomic_permutation import AtomicPermutation
from .coordinate_perturbation import CoordinatePerturbation
from .heredity_mutation import Heredity
from .lattice_strain import LatticeStrain
from .mirror_mutation import MirrorMutation
from .rotational_mutation import RotationalMutation
from .soft_mutation import SoftMutation



# -*- coding: utf-8 -*-

from .incar import Incar
from .kpoints import Kpoints
from .poscar import Poscar
from .potcar import Potcar
from .potcar_mappings import PBE_POTCAR_MAPPINGS, PBE_POTCAR_MAPPINGS_LOW_QUALITY



# -*- coding: utf-8 -*-

# Alternative implementations of this class from ASE and PyMatGen
# https://gitlab.com/ase/ase/-/blob/master/ase/io/espresso.py
# https://github.com/materialsproject/pymatgen/blob/master/pymatgen/io/pwscf.py

import logging
from pathlib import Path

from pymatgen.core import Lattice

from simmate.apps.quantum_espresso.inputs.k_points import Kpoints
from simmate.apps.quantum_espresso.inputs.pwscf_in_modifiers import (
    keyword_modifier_ecutrho,
    keyword_modifier_ecutwfc,
    keyword_modifier_nat,
    keyword_modifier_ntyp,
    keyword_modifier_pseudo_dir,
)
from simmate.toolkit import Structure
from simmate.utilities import str_to_datatype


class PwscfInput:
    """
    Base input file class. Right now, it does not support symmetry and is
    very basic.

    The official input file docs (which this class wraps) is located here:
        https://www.quantum-espresso.org/Doc/INPUT_PW.html

    Currently we make the following assumptions about the file:
        - section titles are in all caps
        - all key-value pairs are on separate lines
        - there are no commas at the end of lines (for key-value pair sections)
        - float values use "e" instead of "d" (ex: 1.23e-4)
        - no "if_pos" values are used in the ATOMIC_POSITIONS section
        - use CELL_PARAMETERS with 'angstrom'. avoid `alat` and `bohr`
        - use ATOMIC_POSITIONS with 'crystal' or 'angstrom'. avoid others
    """

    # This class is a fork of PWInput from pymatgen. Many changes have been
    # made, but it good to know where we start & where we can contrib back to.
    # https://github.com/materialsproject/pymatgen/blob/master/pymatgen/io/pwscf.py

    PARAMETER_SECTIONS = [
        # key-value pairs
        "&CONTROL",
        "&SYSTEM",
        "&ELECTRONS",
        "&IONS",
        "&CELL",
        "&FCP",
        "&RISM",
        # structure-based (each has a unique format)
        "ATOMIC_SPECIES",
        "ATOMIC_POSITIONS",
        "K_POINTS",
        "ADDITIONAL_K_POINTS",
        "CELL_PARAMETERS",
        "CONSTRAINTS",
        "OCCUPATIONS",
        "ATOMIC_VELOCITIES",
        "ATOMIC_FORCES",
        "SOLVENTS",
        "HUBBARD",
    ]

    # establish type mappings for common parameters.
    # Note: this is for ALL sections (control, system, electrons, etc.)
    PARAMETER_MAPPINGS = {
        # STRINGS - these must be quoted in the input file
        "calculation": str,
        "title": str,
        "verbosity": str,
        "restart_mode": str,
        "outdir": str,
        "wfcdir": str,
        "prefix": str,
        "disk_io": str,
        "pseudo_dir": str,
        "occupations": str,
        "smearing": str,
        "pol_type": str,
        "input_dft": str,
        "exxdiv_treatment": str,
        "dmft_prefix": str,
        "constrained_magnetization": str,
        "assume_isolated": str,
        "esm_bc": str,
        "vdw_corr": str,
        "mixing_mode": str,
        "diagonalization": str,
        "efield_phase": str,
        "startingpot": str,
        "startingwfc": str,
        "ion_positions": str,
        "ion_velocities": str,
        "ion_dynamics": str,
        "pot_extrapolation": str,
        "wfc_extrapolation": str,
        "ion_temperature": str,
        "cell_dynamics": str,
        "cell_dofree": str,
        "fcp_dynamics": str,
        "fcp_temperature": str,
        "closure": str,
        "starting1d": str,
        "starting3d": str,
        "laue_wall": str,
        # BOOLEANS
        "wf_collect": bool,
        "tstress": bool,
        "tprnfor": bool,
        "lkpoint_dir": bool,
        "tefield": bool,
        "dipfield": bool,
        "lelfield": bool,
        "lorbm": bool,
        "lberry": bool,
        "lfcpopt": bool,
        "monopole": bool,
        "nosym": bool,
        "nosym_evc": bool,
        "noinv": bool,
        "no_t_rev": bool,
        "force_symmorphic": bool,
        "use_all_frac": bool,
        "one_atom_occupations": bool,
        "starting_spin_angle": bool,
        "noncolin": bool,
        "x_gamma_extrapolation": bool,
        "lda_plus_u": bool,
        "lspinorb": bool,
        "london": bool,
        "ts_vdw_isolated": bool,
        "xdm": bool,
        "uniqueb": bool,
        "rhombohedral": bool,
        "realxz": bool,
        "block": bool,
        "scf_must_converge": bool,
        "adaptive_thr": bool,
        "diago_full_acc": bool,
        "tqr": bool,
        "remove_rigid_rot": bool,
        "refold_pos": bool,
        # FLOATS
        "etot_conv_thr": float,
        "forc_conv_thr": float,
        "conv_thr": float,
        "Hubbard_U": float,
        "Hubbard_J0": float,
        "degauss": float,
        "starting_magnetization": float,
        # INTEGERS
        "nstep": int,
        "iprint": int,
        "nberrycyc": int,
        "gdir": int,
        "nppstr": int,
        "ibrav": int,
        "nat": int,
        "ntyp": int,
        "nbnd": int,
        "nr1": int,
        "nr2": int,
        "nr3": int,
        "nr1s": int,
        "nr2s": int,
        "nr3s": int,
        "nspin": int,
        "nqx1": int,
        "nqx2": int,
        "nqx3": int,
        "lda_plus_u_kind": int,
        "edir": int,
        "report": int,
        "esm_nfit": int,
        "space_group": int,
        "origin_choice": int,
        "electron_maxstep": int,
        "mixing_ndim": int,
        "mixing_fixed_ns": int,
        "ortho_para": int,
        "diago_cg_maxiter": int,
        "diago_david_ndim": int,
        "nraise": int,
        "bfgs_ndim": int,
        "if_pos": int,
        "nks": int,
        "nk1": int,
        "nk2": int,
        "nk3": int,
        "sk1": int,
        "sk2": int,
        "sk3": int,
        "nconstr": int,
    }

    def __init__(
        self,
        structure: Structure,
        kpoints: Kpoints,
        psuedo_mappings: dict = {},
        control: dict = {},
        system: dict = {},
        electrons: dict = {},
        ions: dict = {},
        cell: dict = {},
        fcp: dict = {},
        rism: dict = {},
    ):
        """
        Initializes a PWSCF input file.
        """

        # check for psuedo mappings:
        if not psuedo_mappings:
            logging.warning(
                "No psuedopotential mappings were provided. If you are trying to "
                "use the Simmate defaults (from SSSP), then make sure you have "
                "ran 'simmate-qe setup sssp' to download the files for you. "
                "Otherwise, make sure you provide a mapping and that the "
                "files are present in '~/simmate/quantum_espresso/potentials'"
            )

        self.structure = structure
        self.kpoints = kpoints
        self.psuedo_mappings = psuedo_mappings

        self.control = control
        self.system = system
        self.electrons = electrons
        self.ions = ions
        self.cell = cell
        self.fcp = fcp
        self.rism = rism

    # -------------------------------------------------------------------------

    # Loading methods

    @classmethod
    def from_file(cls, filename: Path | str = "INCAR"):
        """
        Builds an PwscfInput object from a file.
        """
        filename = Path(filename)
        with filename.open() as file:
            content = file.read()
        return cls.from_str(content)

    @classmethod
    def from_str(cls, content: str):
        """
        Builds an PwscfInput object from a string.
        """
        # split the file content into separate lines
        lines = content.split("\n")

        # organize all lines by their section headers
        sections = {}
        current_section = None
        for line in lines:
            # clean the line
            line = line.strip()

            # If the line starts with a # or ! then its a comment and we should skip.
            # It also could be an empty line that we should skip,
            if (
                # empty lines
                not line
                or line.startswith("\n")
                # comment lines
                or line.startswith("#")
                or line.startswith("!")
                # end of section
                or line == "/"
            ):
                continue
            # Check to see if the line is the start of a new section
            # Note, we only grab the first 'word' because you can have headers
            # such as "ATOMIC_POSITIONS angstrom" or "ATOMIC_POSITIONS alat"
            elif line.strip().split()[0] in cls.PARAMETER_SECTIONS:
                current_section = line
                sections[current_section] = []
            else:
                assert current_section is not None  # bug-check
                sections[current_section].append(line)

        # now go through each section and convert/format to as needed
        sections_cleaned = {}
        for section, lines in sections.items():
            # ----------------------

            # these sections are key-value pairs and give a dictionary
            if section in [
                "&CONTROL",
                "&SYSTEM",
                "&ELECTRONS",
                "&IONS",
                "&CELL",
                "&FCP",
                "&RISM",
            ]:
                section_name = section.replace("&", "").lower()
                section_data = {}
                for line in lines:
                    parameter, value = [i.strip() for i in line.split("=")]
                    section_data[parameter] = str_to_datatype(
                        parameter=parameter,
                        value=value,
                        type_mappings=cls.PARAMETER_MAPPINGS,
                        strip_quotes=True,
                    )

            # All other sections need special handling
            # ----------------------

            elif section == "ATOMIC_SPECIES":
                section_name = section.lower()
                section_data = []
                # each line follows...
                #   X(symbol)    Mass_X     PseudoPot_X
                # for ex:
                #   Si  28.086  Si.pz-vbc.UPF
                # Note: mass is only used for MD calcs.
                for line in lines:
                    element, mass, psuedo_file = line.strip().split()
                    # TODO: use an Element base class...?
                    specie = {
                        "element": element.strip(),
                        "mass": mass.strip(),
                        "psuedo_file": psuedo_file.strip(),
                    }
                    section_data.append(specie)

            # ----------------------

            elif section.startswith("ATOMIC_POSITIONS"):
                section_name, section_mode = section.lower().split()
                section_data = {
                    "mode": section_mode,
                    "data": [],
                }
                # each line follows...
                #   X(symbol)    x    y    z
                # for ex:
                #   Si 0.00 0.00 0.00
                #   Si 0.25 0.25 0.25
                # Note: mass is only used for MD calcs.
                for line in lines:
                    # TODO: use a PeriodicSite base class...?
                    element, x, y, z = line.strip().split()
                    specie = {
                        "element": element.strip(),
                        "coords": [float(x), float(y), float(z)],
                    }
                    section_data["data"].append(specie)

            # ----------------------

            elif section.startswith("K_POINTS") or section.startswith(
                "ADDITIONAL_K_POINTS"
            ):
                section_name, *section_mode = section.lower().split()
                if not section_mode:
                    section_mode = "tpiba"  # using default value
                    # Note: if the is NO section at all, then the default is
                    # actually gamma. This is handled within `from_dict`
                else:
                    section_mode = section_mode[0]
                section_data = {
                    "mode": section_mode,
                    "data": [],
                }
                if section_mode == "gamma":
                    section_data.pop("data")  # no extra info needed
                elif section_mode == "automatic":
                    # there should be a single line for this mode.
                    #   nk1  nk2  nk3  sk1  sk2  sk3
                    # n = grid, s = offset
                    assert len(lines) == 1
                    nk1, nk2, nk3, sk1, sk2, sk3 = lines[0].strip().split()
                    section_data["data"] = {
                        "grid": [int(nk1), int(nk2), int(nk3)],
                        "offset": [int(sk1), int(sk2), int(sk3)],
                    }

                elif section_mode in [
                    "tpiba",
                    "crystal",
                    "tpiba_b",
                    "crystal_b",
                    "tpiba_c",
                    "crystal_c",
                ]:
                    # The first line is always just the total number of kpts.
                    # nkpts = line[0]  # we don't store this because it can be inferred

                    # Then each line follows...
                    #   xk_x(1) 	 xk_y(1) 	 xk_z(1) 	 wk(1)
                    # for ex:
                    #   10
                    #   0.1250000  0.1250000  0.1250000   1.00
                    #   0.1250000  0.1250000  0.3750000   3.00
                    #   ..... (+ 8 more lines for kpts)
                    for line in lines[1:]:
                        x, y, z, weight = line.strip().split()
                        kpt = {
                            "coords": [float(x), float(y), float(z)],
                            "weight": float(weight),
                        }
                        section_data["data"].append(kpt)

            # ----------------------

            elif section.startswith("CELL_PARAMETERS"):
                section_name, section_mode = section.lower().split()
                section_data = {
                    "mode": section_mode,
                    "data": [],
                }
                # section format follows...
                #   v1(1)  v1(2)  v1(3)    ... 1st lattice vector
                #   v2(1)  v2(2)  v2(3)    ... 2nd lattice vector
                #   v3(1)  v3(2)  v3(3)    ... 3rd lattice vector
                assert len(lines) == 3  # bug-check
                lattice = [[float(v.strip()) for v in l.split()] for l in lines]
                section_data["data"].append(lattice)

            # ----------------------

            elif section == "CONSTRAINTS":
                raise NotImplementedError()  # TODO

            elif section == "OCCUPATIONS":
                raise NotImplementedError()  # TODO

            elif section.startswith("ATOMIC_VELOCITIES"):
                raise NotImplementedError()  # TODO

            elif section == "ATOMIC_FORCES":
                raise NotImplementedError()  # TODO

            elif section.startswith("SOLVENTS"):
                raise NotImplementedError()  # TODO

            elif section.startswith("HUBBARD"):
                raise NotImplementedError()  # TODO

            # ----------------------

            # regardless of the method above, we now have cleaned the section
            sections_cleaned[section_name] = section_data

        # we now have all the infomation we need in a clean dictionary format
        return cls.from_dict(sections_cleaned)

    @classmethod
    def from_dict(cls, data: dict):
        # We do not yet support symmetry-based inputs for lattice/structures
        lattice_symmetry = data["system"]["ibrav"]  # ibrav is a required input
        if lattice_symmetry != 0:
            raise NotImplementedError("PwscfInput currently only supports ibrav=0")
            # TODO: add support via Structure.from_spacegroup

        # ----------------------

        # build the lattice

        lattice_parameters = data.get("cell_parameters")
        if not lattice_parameters:
            raise Exception(
                "PwscfInput requires cell parameters to determine the structure"
            )

        # regardless of input, we convert to Angstrom for simmate
        lattice_units = lattice_parameters.get("mode")
        if lattice_units == "alat":
            raise NotImplementedError("alat is not yet supported for cell parameters")
        elif lattice_units == "bohr":
            raise NotImplementedError("bohr is not yet supported for cell parameters")
        elif lattice_units == "angstrom":
            # we already have the matrix in proper units
            lattice_matrix = lattice_parameters["data"]
        else:
            raise Exception(f"Unknown cell parameters mode: {lattice_units}.")

        lattice = Lattice(matrix=lattice_matrix)

        # ----------------------

        # build the structure using lattice + atomic sites

        site_parameters = data.get("atomic_positions")
        if not site_parameters:
            raise Exception(
                "PwscfInput requires atomic positions to determine the structure"
            )

        # regardless of input, we convert to either Angstrom or fractional coords
        site_units = site_parameters.get("mode")
        if site_units == "alat":
            raise NotImplementedError("alat is not yet supported for cell parameters.")
        elif site_units == "bohr":
            raise NotImplementedError("bohr is not yet supported for cell parameters.")
        elif site_units == "angstrom":
            coords_are_cartesian = True
            lattice_matrix = lattice_parameters["data"]
        elif site_units == "crystal":
            coords_are_cartesian = False
        elif site_units == "crystal_sg":
            raise NotImplementedError(
                "crystal_sg is not yet supported for cell parameters."
            )
        else:
            raise Exception(f"Unknown atomic positions mode: {lattice_units}.")

        # regardless of mode, we bring all site coords and elements into
        # independent lists
        species = [site["element"] for site in site_parameters["data"]]
        coords = [site["coords"] for site in site_parameters["data"]]

        structure = Structure(
            lattice=lattice,
            species=species,
            coords=coords,
            coords_are_cartesian=coords_are_cartesian,
        )

        # ----------------------

        # build k-points info
        # TODO: have base class to help here + merge with VASP

        kpt_parameters = data.get("k_points", {})
        # default when no section at all is a single point gamma kpt
        kpoints_mode = kpt_parameters.get("mode", "gamma")

        if kpoints_mode == "gamma":  # Use only the Gamma point.
            kpoints_grid = None
            kpoints_offset = None
            kpoints_weights = None
        elif (
            kpoints_mode == "automatic"
        ):  # Automatically generate a Monkhorst-Pack grid
            kpoints_grid = kpt_parameters["data"]["grid"]
            kpoints_offset = kpt_parameters["data"]["offset"]
            kpoints_weights = None
        elif kpoints_mode in [
            "tpiba",  # Monkhorst-Pack grid specified in reciprocal lattice units
            "crystal",  # Specify k-points in crystal coordinates.
        ]:
            kpoints_offset = None
            kpoints_grid = [k["coords"] for k in kpt_parameters["data"]]
            kpoints_weights = [k["weight"] for k in kpt_parameters["data"]]
        elif kpoints_mode in [
            "tpiba_b",  # Monkhorst-Pack grid specified in reciprocal lattice units, with a shift.
            "tpiba_c",  #  Monkhorst-Pack grid specified in reciprocal lattice units, with a different shift.
            "crystal_b",  # Specify k-points in crystal coordinates with a shift.
            "crystal_c",  # Specify k-points in crystal coordinates with a different shift.
        ]:
            raise NotImplementedError(
                f"{kpoints_mode} is not yet supported for k points"
            )
        else:
            raise Exception(f"Unknown k points mode: {kpoints_mode}.")

        kpoints = Kpoints(
            mode=kpoints_mode,
            grid=kpoints_grid,
            offset=kpoints_offset,
            weights=kpoints_weights,
        )

        # ----------------------

        return cls(
            structure=structure,
            kpoints=kpoints,
            control=data.get("control", {}),  # !!! should I pop psuedo_file?
            system=data.get("system", {}),
            electrons=data.get("electrons", {}),
            ions=data.get("ions", {}),
            cell=data.get("cell", {}),
            fcp=data.get("fcp", {}),
            rism=data.get("rism", {}),
        )

    # -------------------------------------------------------------------------

    # Export methods

    def to_str(self) -> str:
        final_str = ""

        # ----------------------

        # place key-value sections up top
        # Note: we pull the final evaluated settings, which fills out modifiers
        # such as "__per_atom" or "__auto__"

        for section, content in self.evaluated_settings.items():
            if not content:
                continue

            # section title
            final_str += f"&{section.upper()}\n"
            # parameters (we copy because we modify values some)
            for key, value in content.copy().items():
                # if the input type is a str, we want to make sure it's
                # wrapped in single quotes for pw-scf to read
                if self.PARAMETER_MAPPINGS.get(key) == str:
                    value = f"'{value}'"
                # booleans are written differently than in python
                elif self.PARAMETER_MAPPINGS.get(key) == bool:
                    value = f".{str(value).lower()}."
                final_str += f"\t{key} = {value}\n"
            # section ending
            final_str += "/\n\n"

        # ----------------------

        # then write structural information

        # lattice info
        # string is the same as str(lattice) but we just indent the entire thing
        final_str += "CELL_PARAMETERS angstrom\n"
        lattice_str = str(self.structure.lattice).replace("\n", "\n ")
        final_str += f" {lattice_str}\n\n"

        # specie info
        final_str += "ATOMIC_SPECIES\n"
        for element in self.structure.composition:
            psuedo_name = self.psuedo_mappings[element.symbol]["filename"]
            final_str += (
                f" {element.symbol}  {float(element.atomic_mass)}  {psuedo_name}\n"
            )
        final_str += "\n"  # extra empty line after final specie

        # site info
        # ATOMIC_POSITIONS (angstrom or crystal)
        site_mode = "crystal"  # !!! we assume frac coords for now
        final_str += f"ATOMIC_POSITIONS {site_mode}\n"
        for site in self.structure:
            final_str += f" {site.specie.symbol} {site.a} {site.b} {site.c}\n"
        final_str += "\n"  # extra empty line after final site

        # ----------------------

        # then write kpoint information

        # we evaluate the class in case there are modifiers like __kpt_density
        kpoints = self.evaluated_k_points

        final_str += f"K_POINTS {kpoints.mode}\n"
        if kpoints.mode == "gamma":
            pass  # no extra lines needed

        elif kpoints.mode == "automatic":
            # a single line is needed
            nk1, nk2, nk3 = kpoints.grid
            sk1, sk2, sk3 = kpoints.offset
            final_str += f" {nk1} {nk2} {nk3} {sk1} {sk2} {sk3}\n"

        elif kpoints.mode in ["tpiba", "crystal"]:
            final_str += f" {len(self.kpoints.grid)}\n"
            for kpt, kpt_wt in zip(kpoints.grid, kpoints.weights):
                final_str += f" {kpt[0]} {kpt[1]} {kpt[2]} {kpt_wt}\n"

        elif kpoints.mode in [
            "tpiba_b",
            "tpiba_c",
            "crystal_b",
            "crystal_c",
        ]:
            raise NotImplementedError(
                f"{kpoints.mode} is not yet supported for k points"
            )

        else:
            raise Exception(f"Unknown k points mode: {kpoints.mode}.")

        final_str += "\n"

        # ----------------------

        return final_str

    def to_file(self, filename: Path | str = "pwscf.in"):
        """
        Write PWSCF input settings to a file.
        Args:
            filename (str): filename to write to.
        """
        # we just take the string format and put it in a file
        filename = Path(filename)
        with filename.open("w") as file:
            file.write(self.to_str())

    # -------------------------------------------------------------------------

    @property
    def evaluated_settings(self) -> dict:
        final_settings = {}

        for section in [
            "control",
            "system",
            "electrons",
            "ions",
            "cell",
            "fcp",
            "rism",
        ]:
            content = getattr(self, section)
            if content:
                final_settings[section] = self._evaluate_dict(content)

        return final_settings

    def _evaluate_dict(self, data: dict) -> dict:
        # TODO: Refactor and combine this functionality with vasp.Incar

        # First we need to iterate through all parameters and check if we have
        # ones that are structure-specific. For example, we would need to
        # evaluate "ecutwfc__per_atom". We go through all these and collect the
        # parameters into a final settings list.
        final_settings = {}
        for parameter, value in data.items():
            # if there is no modifier attached to the parameter, we just keep it as-is
            if "__" not in parameter:
                final_settings[parameter] = value

            # Otherwise we have a modifier like "__density" and need to evaluate it
            else:
                # make sure we have a structure supplied because all modifiers
                # require one.
                if not self.structure:
                    raise Exception(
                        "It looks like you used a keyword modifier but didn't "
                        f"supply a structure! If you want to use {parameter}, "
                        "then you need to make sure you provide a structure so "
                        "that the modifier can be evaluated."
                    )

                # separate the input into the base parameter and modifier.
                # This also overwrites what our paramter value is.
                parameter, modifier_tag = parameter.split("__")

                # check that this class has this modifier supported. It should
                # be a method named "keyword_modifier_mymodifier".
                # If everything looks good, we grab the modifier function.
                # If the tag is just "auto", then we look for a tag that is
                # named after the input variable
                if modifier_tag == "auto":
                    modifier_fxn_name = "keyword_modifier_" + parameter
                else:
                    modifier_fxn_name = "keyword_modifier_" + modifier_tag
                if hasattr(self, modifier_fxn_name):
                    modifier_fxn = getattr(self, modifier_fxn_name)
                else:
                    raise AttributeError(
                        """
                        It looks like you used a keyword modifier that hasn't
                        been defined yet! If you want something like `ENCUT__per_atom`,
                        then you need to make sure there is a `keyword_modifier_per_atom`
                        method available. "auto" modifiers are a special case where,
                        `ENCUT__auto` needs the method to be `keyword_modifier_ENCUT`.
                        """
                    )

                # now that we have the modifier function, let's use it to update
                # our value for this keyword.
                value = modifier_fxn(self.structure, value)

                # sometimes the modifier returns None. In this case we don't
                # set anything in the INCAR, but leave it to the programs
                # default value.
                if not value:
                    continue
                # BUG: Are there cases where None is return but we still want
                # to write it to the INCAR? If so, it'd be skipped here.

                # if the "parameter" is actually "multiple_keywords", then we
                # have our actual parameters as a dictionary. We need to
                # pull these out of the "value" we have.
                if parameter == "multiple_keywords":
                    for subparameter, subvalue in value.items():
                        final_settings[subparameter] = subvalue

                # otherwise we were just given back an update value
                else:
                    final_settings[parameter] = value
        return final_settings

    @property
    def evaluated_k_points(self) -> Kpoints:
        # TODO: move this to the Kpoints class
        return Kpoints(**self._evaluate_dict(self.kpoints.to_dict()))

    @classmethod
    def add_keyword_modifier(cls, keyword_modifier: callable):
        # !!! this is a copy/paste of the method from vasp.Incar
        if hasattr(cls, keyword_modifier.__name__):
            raise Exception(
                "The Incar class already has a modifier with this name. "
                "Please use a different modifier name to avoid conflicts."
            )
        setattr(cls, keyword_modifier.__name__, staticmethod(keyword_modifier))


# set some default keyword modifiers
for modifier in [
    keyword_modifier_pseudo_dir,
    keyword_modifier_nat,
    keyword_modifier_ntyp,
    keyword_modifier_ecutwfc,
    keyword_modifier_ecutrho,
]:
    PwscfInput.add_keyword_modifier(modifier)



# -*- coding: utf-8 -*-

import pytest
from pandas import DataFrame

from simmate.database.base_data_types import StaticEnergy
from simmate.toolkit import Structure


@pytest.mark.django_db
def test_static_energy_table(structure, tmp_path):
    # test writing columns
    StaticEnergy.show_columns()

    # test writing to database
    structure_db = StaticEnergy.from_run_context(
        run_id="example-id-123",
        workflow_name="example.test.workflow",
        structure=structure,
        workflow_version="1.2.3",
    )
    structure_db.save()

    # try grabbing the calculation again and make sure it loaded from the
    # database rather than creating a new entry
    structure_db2 = StaticEnergy.from_run_context(
        run_id="example-id-123",
        workflow_name="example.test.workflow",
        workflow_version="1.2.3",
    )
    assert structure_db.id == structure_db2.id

    # test converting back to toolkit and ensuring the structure is the
    # same as before
    structure_new = structure_db.to_toolkit()
    assert structure == structure_new

    # test converting search results to dataframe and to toolkit
    df = StaticEnergy.objects.to_dataframe()
    assert isinstance(df, DataFrame)
    structures = StaticEnergy.objects.to_toolkit()
    assert isinstance(structures, list)
    assert isinstance(structures[0], Structure)

    # update the database entry using a Vasprun result.
    # from pymatgen.io.vasp.outputs import Vasprun
    # TODO:

    # test writing and reloading these from and archive
    archive_filename = tmp_path / "archive.zip"
    StaticEnergy.objects.to_archive(archive_filename)
    StaticEnergy.load_archive(
        archive_filename,
        confirm_override=True,
        delete_on_completion=True,
    )



# -*- coding: utf-8 -*-

from pymatgen.io.vasp.inputs import Kpoints
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.bandstructure import HighSymmKpath


def get_hse_kpoints(
    structure,
    added_kpoints: list = [],
    mode="line",  # or uniform
    reciprocal_density: float = 50,
    kpoints_line_density: float = 20,
):
    # TODO: move these options up to setup() so they can be set in the yaml file
    # BUGFIX:
    #   https://github.com/jacksund/simmate/issues/388
    # This code is copied and modified from...
    #   https://github.com/materialsproject/pymatgen/blob/b4fb2593dc0d0f1d2de7e3bb88ed0b0e1d039733/pymatgen/io/vasp/sets.py#L1376-L1421
    kpts: list[int | float | None] = []
    weights: list[float | None] = []
    all_labels: list[str | None] = []

    # for both modes, include the Uniform mesh w/standard weights
    grid = Kpoints.automatic_density_by_vol(structure, reciprocal_density).kpts
    ir_kpts = SpacegroupAnalyzer(structure, symprec=0.1).get_ir_reciprocal_mesh(grid[0])
    for k in ir_kpts:
        kpts.append(k[0])
        weights.append(int(k[1]))
        all_labels.append(None)

    # for both modes, include any user-added kpoints w/zero weight
    for k in added_kpoints:
        kpts.append(k)
        weights.append(0.0)
        all_labels.append("user-defined")

    # for line mode only, add the symmetry lines w/zero weight
    if mode.lower() == "line":
        kpath = HighSymmKpath(structure)
        frac_k_points, labels = kpath.get_kpoints(
            line_density=kpoints_line_density,
            coords_are_cartesian=False,
        )

        for k, f in enumerate(frac_k_points):
            kpts.append(f)
            weights.append(0.0)
            all_labels.append(labels[k])

    comment = (
        "HSE run along symmetry lines"
        if mode.lower() == "line"
        else "HSE run on uniform grid"
    )

    return Kpoints(
        comment=comment,
        style=Kpoints.supported_modes.Reciprocal,
        num_kpts=len(kpts),
        kpts=kpts,  # type: ignore
        kpts_weights=weights,
        labels=all_labels,
    )



# -*- coding: utf-8 -*-

from simmate.apps.vasp.inputs import PBE_POTCAR_MAPPINGS
from simmate.apps.vasp.workflows.base import VaspWorkflow
from simmate.apps.vasp.workflows.diffusion import (
    NebAllPathsWorkflow,
    SinglePathWorkflow,
    VaspNebFromImagesWorkflow,
)

# -----------------------------------------------------------------------------

# BULK UNITCELL RELAXATION


class Relaxation__Vasp__WarrenLabQuick(VaspWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    _incar = dict(
        ALGO="Fast",
        EDIFF=1e-05,
        ENCUT=520,  # TODO: set dynamically to be 1.3x highest elemental ENMAX
        IBRION=2,
        ICHARG=1,
        ISIF=3,
        ISPIN=2,
        ISYM=0,
        IVDW=12,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=200,
        NSW=99,
        PREC="Accurate",
        ISMEAR=0,
        SIGMA=0.05,
        KSPACING=0.5,
        LMAXMIX=4,
    )
    error_handlers = []


# -----------------------------------------------------------------------------

# BULK UNITCELL STATIC ENERGY


class StaticEnergy__Vasp__WarrenLabQuick(Relaxation__Vasp__WarrenLabQuick):
    _incar_updates = dict(IBRION=-1, NSW=0)


# -----------------------------------------------------------------------------

# ENDPOINT SUPERCELL RELAXATIONS


class Relaxation__Vasp__WarrenLabQuickNebEndpoint(VaspWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    _incar = dict(
        ALGO="Fast",
        EDIFF=5e-04,
        EDIFFG=-0.02,
        ENCUT=520,  # TODO: set dynamically to be 1.3x highest elemental ENMAX
        IBRION=2,
        ICHARG=1,
        ISIF=2,
        ISPIN=2,
        ISYM=0,
        IVDW=12,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        LCHARG=False,
        NELM=200,
        NSW=99,
        PREC="Accurate",
        ISMEAR=0,
        SIGMA=0.05,
        KSPACING=0.5,
        LMAXMIX=4,
    )
    error_handlers = []


# -----------------------------------------------------------------------------

# ENDPOINT SUPERCELL STATIC ENERGY


class StaticEnergy__Vasp__WarrenLabQuickNebEndpoint(
    Relaxation__Vasp__WarrenLabQuickNebEndpoint
):
    _incar_updates = dict(IBRION=-1, NSW=0)


# -----------------------------------------------------------------------------

# NEB FROM IMAGES


class Diffusion__Vasp__WarrenLabQuickCiNebFromImages(VaspNebFromImagesWorkflow):
    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS
    _incar = dict(
        ALGO="Fast",
        EDIFF=5e-04,
        EDIFFG=-0.02,
        ENCUT=520,  # TODO: set dynamically to be 1.3x highest elemental ENMAX
        IBRION=3,
        ICHARG=1,
        ISIF=2,
        ISPIN=2,
        ISYM=0,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        LCHARG=False,
        NELM=200,
        NSW=99,
        PREC="Accurate",
        ISMEAR=0,
        SIGMA=0.05,
        KSPACING=0.5,
        LMAXMIX=4,
        IMAGES=3,
    )
    error_handlers = []


# -----------------------------------------------------------------------------

# SINGLE PATH NEB


class Diffusion__Vasp__WarrenLabQuickNebSinglePath(SinglePathWorkflow):
    endpoint_relaxation_workflow = Relaxation__Vasp__WarrenLabQuickNebEndpoint
    endpoint_energy_workflow = StaticEnergy__Vasp__WarrenLabQuickNebEndpoint
    from_images_workflow = Diffusion__Vasp__WarrenLabQuickCiNebFromImages


# -----------------------------------------------------------------------------

# ALL PATHS NEB


class Diffusion__Vasp__NebAllPathsWarrenLabQuick(NebAllPathsWorkflow):
    bulk_relaxation_workflow = Relaxation__Vasp__WarrenLabQuick
    bulk_static_energy_workflow = StaticEnergy__Vasp__WarrenLabQuick
    single_path_workflow = Diffusion__Vasp__WarrenLabQuickNebSinglePath


# -----------------------------------------------------------------------------



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate database. All commands are 
accessible through the `simmate database` command.
"""

from pathlib import Path

import typer

database_app = typer.Typer(rich_markup_mode="markdown")


@database_app.callback(no_args_is_help=True)
def base_command():
    """A group of commands for managing your database"""
    pass


@database_app.command()
def reset(confirm_delete: bool = False, use_prebuilt: bool = None):
    """
    Removes any existing data and sets up a clean database


    - `--confirm-delete`: automatically confirms you want to delete the
    existing database.
    :warning::warning: Use this with caution.:warning::warning:

    - `--use-prebuilt` and `--no-use-prebuilt`: automatically say yes/no to a
    prebuilt database. This only applies if you are using sqlite.
    """
    from simmate.configuration import settings

    database_name = str(settings.database.name)
    print(f"\nUsing {database_name}")

    # make sure the user knows what they are doing and actually wants to continue
    if not confirm_delete:
        typer.confirm(
            "\nWARNING: This deletes your current database and cannot be undone. \n"
            "We highly recommend you make a copy of your database before doing this. \n\n"
            "Do you still want to continue?",
            abort=True,
        )

    # if the user has a SQLite3 backend, ask if they'd like to use a prebuilt
    # database to begin
    if settings.database_backend == "sqlite3" and use_prebuilt is None:
        use_prebuilt = typer.confirm(
            "\nIt looks like you are using the default database backend (sqlite3). \n"
            "Would you like to use a prebuilt-database with all third-party data "
            "already loaded? \n"
            "If this is the first time you using the prebuild, this will "
            "involve a ~1.5GB \ndownload and will unpack to roughly 5GB.\n\n"
            "We recommend answering 'yes' for beginners."
        )

    from simmate.database import connect
    from simmate.database.utilities import reset_database

    reset_database(use_prebuilt=use_prebuilt)


@database_app.command()
def update():
    """Updates the database with any changes made"""

    from simmate.database import connect
    from simmate.database.utilities import update_database

    update_database()


@database_app.command()
def dump_data(filename: Path = "database_dump.json", exclude: list[str] = []):
    """
    Takes the Simmate database and writes it to a json file

    - `--filename` is the file to write the all the JSON data to
    """

    from simmate.database import connect
    from simmate.database.utilities import dump_database_to_json

    dump_database_to_json(filename=filename, exclude=exclude)


@database_app.command()
def load_data(filename: Path = "database_dump.json"):
    """
    Takes a JSON database and loads it into the Simmate database

    - `--filename` is the file to load the all the JSON data from
    """

    from simmate.database import connect
    from simmate.database.utilities import load_database_from_json

    load_database_from_json(filename=filename)


@database_app.command()
def load_remote_archives(parallel: bool = False):
    """
    Loads all third-party data into the database

    This utility helps with initializing a new database build.

    :WARNING:
    This can take several hours to run and there is no pause/continuation
    implemented. This runs substantially faster when you are using a cloud
    database backend (e.g. Postgres) and use the `--parallel` flag.

    If you are using SQLite, we highly recommend using loading the prebuild
    when you run `simmate database reset` instead.

    - `--parallel`: (not allowed for SQLite3) loads data using parallel processing
    """

    from simmate.database import connect
    from simmate.database.third_parties import load_remote_archives

    load_remote_archives(parallel=parallel)



# -*- coding: utf-8 -*-

from pathlib import Path

import typer

inputs_app = typer.Typer(rich_markup_mode="markdown")


@inputs_app.callback(no_args_is_help=True)
def inputs():
    """
    A group of commands for creating and analyzing VASP inputs
    """
    pass


@inputs_app.command()
def get_potcar(elements: list[str], functional: str = "PBE", combine: bool = False):
    """
    Grabs the POTCAR file(s) of the requested type

    ex: `simmate-vasp inputs get-potcar Y C F`
    """
    from pymatgen.core import Element

    from simmate.apps.vasp.inputs import Potcar

    elements = [Element(e) for e in elements]
    for element in elements:
        Potcar.to_file_from_type(
            elements=[element],
            functional=functional,
            filename=f"POTCAR_{element}",
        )

    if combine and len(elements) > 1:
        Potcar.to_file_from_type(elements, functional)


@inputs_app.command()
def compare_incars(incar1: Path, incar2: Path):
    """
    Compares settings between two INCAR files and shows their similarities and
    differences
    """
    import yaml

    from simmate.apps.vasp.inputs import Incar

    incar1 = Incar.from_file(filename=incar1)
    incar2 = Incar.from_file(filename=incar2)

    result = incar1.compare_incars(incar2)

    # BUG-FIX: Tuples print ugly, so we convert them to a list
    for key, value in result["Different"].items():
        if isinstance(value, tuple):
            result["Different"][key] = list(value)

    print(yaml.dump(result))



# -*- coding: utf-8 -*-

from ase.ga.standardmutations import MirrorMutation as ASEMirrorMutation
from ase.ga.utilities import closest_distances_generator
from pymatgen.io.ase import AseAtomsAdaptor

from simmate.toolkit import Structure
from simmate.toolkit.transformations.base import Transformation


class MirrorMutation(Transformation):
    """
    This is a wrapper around the `MirrorMutation` in ase.ga
    https://gitlab.com/ase/ase/-/blob/master/ase/ga/standardmutations.py
    """

    name = "from_ase.MirrorMutation"
    io_scale = "one_to_one"
    ninput = 1
    allow_parallel = False

    @staticmethod
    def apply_transformation(
        structure: Structure,
        ratio_of_covalent_radii: float = 0.1,
    ) -> Structure:
        #!!! TO-DO. In many cases, you can perform this operation and simply
        # get back the original structure. I should check and make sure
        # that I'm actually returning a new structure

        # ----------- SETUP (consider caching as class attribute) -------------
        # the closest_distances_generator is exactly the same as an
        # element-dependent distance matrix expect ASE puts this in dictionary
        # form the function requires a list of element integers
        element_ints = [element.number for element in structure.composition]
        # the default of the ratio of covalent radii (0.1) is based on the
        # ASE tutorial of this function
        element_distance_matrix = closest_distances_generator(
            element_ints, ratio_of_covalent_radii
        )
        # ---------------------------------------------------------------------

        # first I need to convert the structures to an ASE atoms object
        structure_ase = AseAtomsAdaptor.get_atoms(structure)

        # now we can make the generator
        mirror = ASEMirrorMutation(
            blmin=element_distance_matrix,  # distance cutoff matrix
            n_top=int(
                structure.composition.num_atoms
            ),  # number of atoms to optimize. I set this to all
            # reflect=False, #!!! need to look into this more
            # rng=np.random,
            # verbose=False
        )

        #!!! Their code suggests the use of .get_new_individual() but I
        # think .mutate() is what we'd like
        new_structure_ase = mirror.mutate(structure_ase)

        # if the mutation fails, None is return
        if not new_structure_ase:
            return False

        # if it was successful, we have a new Atoms object
        # now convert back to a pymatgen object
        new_structure = AseAtomsAdaptor.get_structure(new_structure_ase)

        return new_structure



# -*- coding: utf-8 -*-

import itertools
import logging
import time
from pathlib import Path

import numpy
from pymatgen.analysis.reaction_calculator import Reaction, ReactionError
from rich.progress import track

from simmate.apps.evolution.models.chemical_system import ChemicalSystemSearch
from simmate.apps.evolution.singleshot_sources.known import get_known_structures
from simmate.apps.evolution.singleshot_sources.prototypes import (
    get_structures_from_prototypes,
)
from simmate.apps.evolution.workflows.fixed_composition import (
    StructurePrediction__Toolkit__FixedComposition,
)
from simmate.apps.evolution.workflows.utilities import write_and_submit_structures
from simmate.engine import Workflow
from simmate.toolkit import Composition
from simmate.workflows.utilities import get_workflow


class StructurePrediction__Toolkit__ChemicalSystem(Workflow):
    """
    Runs an evolutionary search algorithm to predict the most stable structures
    of a binary phase system (e.g Na-Cl or Y-C)
    """

    description_doc_short = "hull diagram for a two-element system (e.g. Na-Cl)"

    database_table = ChemicalSystemSearch

    fixed_comp_workflow = StructurePrediction__Toolkit__FixedComposition

    @classmethod
    def run_config(
        cls,
        chemical_system: str,
        max_atoms: int = 10,
        subworkflow_name: str = "relaxation.vasp.staged",
        subworkflow_kwargs: dict = {},
        max_stoich_factor: int = 4,
        nfirst_generation: int = 15,
        nsteadystate: int = 40,
        sleep_step: float = 300,
        directory: Path = None,
        singleshot_sources: list[str] = [
            "third_parties",
            "prototypes",
        ],
        run_id: str = None,
        **kwargs,
    ):
        # ---------------------------------------------------------------------
        # Setting up
        # ---------------------------------------------------------------------

        # grab the calculation table linked to this workflow run
        search_datatable = cls.database_table.objects.get(run_id=run_id)

        subworkflow = get_workflow(subworkflow_name)

        # Grab the two elements that we are working with. Also if this raises
        # an error, we want to provide useful feedback to the user
        try:
            endpoint_compositions = [
                Composition(sys) for sys in chemical_system.split("-")
            ]
        except:
            raise Exception(
                "Failed to split {} into elements. Your format for the input "
                "chemical system should be two elements and follow the format "
                "Element1-Element2. Ex: Na-Cl, Ca-N, Y-C, etc."
            )

        # grab all unique elements
        elements = []
        for comp in endpoint_compositions:
            for element in comp.elements:
                if element not in elements:
                    elements.append(element)

        # Find all unique compositions for the list of elements.
        # Note, for complex compositions, this will contain extra compositions
        # that we will want removed. This is addressed in the next step.
        compositions = []
        for nsites in range(1, max_atoms + 1):
            for combo in itertools.combinations_with_replacement(elements, nsites):
                counts = {e.symbol: combo.count(e) for e in elements}
                composition = Composition(**counts)
                compositions.append(composition)

        # For complex chemical systems (e.g. the Sc2C-Sc2CF2 or Y6S4-YF3-Y systems)
        # we need to make sure the composition is contained with that system.
        # The general rule is that it must be possible to stoichiometrically
        # balance the reaction of endpoint compositions with the composition
        # being considered. For example, in the Y6S4-YF3-Y system,
        # Y2SF would be accepted (0.5 Y3S2 + 0.33 YF3 + 0.166 Y -> Y2SF)
        # while YSF would be rejected (0.5 Y3S2 + 0.33 YF3 - 0.833 Y ->  + YSF)
        # (for YSF, note the negative sign for Y)
        compositions_cleaned = []
        for composition in compositions:
            try:
                reaction = Reaction(
                    reactants=endpoint_compositions,
                    products=[composition],
                )
            except ReactionError:
                # reaction could not be balanced and is therefore invalid
                continue

            # iterate and make sure our target composition is close to 1 while
            # all other products are close to 0. Note the is-close checks
            # are because of rounding bugs in the pymatgen code
            is_valid = True
            for product in reaction.products:
                coeff = reaction.get_coeff(product)
                if product == composition and not numpy.isclose(coeff, 1):
                    is_valid = False
                elif product != composition and not numpy.isclose(coeff, 0):
                    is_valid = False
            if is_valid:
                compositions_cleaned.append(composition)
        compositions = compositions_cleaned

        logging.info(f"{len(compositions)} unique compositions will be explored")

        # now condense this list down to just the reduced formulas
        compositions_reduced = []
        for composition in compositions:
            reduced = composition.reduced_composition
            if reduced not in compositions_reduced:
                compositions_reduced.append(reduced)
        logging.info(
            f"This corresponds to {len(compositions_reduced)} reduced compositions"
        )

        # it is also useful to store the "max compositions". This the composition
        # for when the max number of possible sites are used. For example,
        # Ca2N with max atoms of 20 would give a max composition of Ca12N6 (18 atoms).
        # We make this list using only the reduced compositions as input.
        compositions_maxed = []
        for composition in compositions_reduced:
            max_factor = int(max_atoms // composition.num_atoms)
            # BUG: we need to call Composition() to convert the pymatgen
            # object back into a simmate composition.
            max_composition = Composition(max_factor * composition)

            if max_composition not in compositions_maxed:
                compositions_maxed.append(max_composition)

        # ---------------------------------------------------------------------
        # Submitting known structures
        # ---------------------------------------------------------------------

        if "third_parties" in singleshot_sources:
            logging.info("Generating input structures from third-party databases")
            structures_known = []
            for composition in track(compositions_maxed):
                new_structures = get_known_structures(
                    composition,
                    allow_multiples=True,
                    remove_matching=True,
                    nsites__lte=max_atoms,
                )
                structures_known += new_structures
            logging.info(
                f"Generated {len(structures_known)} structures from other databases"
            )
            states_known = write_and_submit_structures(
                structures=structures_known,
                foldername=directory / "from_third_parties",
                workflow=subworkflow,
                workflow_kwargs=subworkflow_kwargs,
            )

        # ---------------------------------------------------------------------
        # Submitting structures from prototypes
        # ---------------------------------------------------------------------

        if "prototypes" in singleshot_sources:
            # Start by generating the singleshot sources for each factor size.
            logging.info("Generating input structures from prototypes")
            structures_prototype = []
            # for singleshot_source in singleshot_sources: ## TODO
            for composition in track(compositions_maxed):
                # generate all prototypes
                new_structures = get_structures_from_prototypes(
                    composition,
                    max_sites=max_atoms,
                    remove_matching=True,
                )
                structures_prototype += new_structures
            logging.info(
                f"Generated {len(structures_prototype)} structures from prototypes"
            )
            states_prototype = write_and_submit_structures(
                structures=structures_prototype,
                foldername=directory / "from_prototypes",
                workflow=subworkflow,
                workflow_kwargs=subworkflow_kwargs,
            )

        # ---------------------------------------------------------------------
        # Wait for singlshot submissions if there are many of them
        # ---------------------------------------------------------------------

        all_submissions = states_prototype + states_known
        if len(all_submissions) > (nsteadystate * 2):
            number_to_wait_for = len(all_submissions) - nsteadystate - 20
            number_to_resume_on = nsteadystate + 20
            logging.info(
                f"Waiting for at least {number_to_wait_for} singleshot "
                "submissions to finish."
            )
            done_waiting = False
            while not done_waiting:
                # go through submitted calculations and count the number of pending
                # calculations
                total_pending = 0
                for state in all_submissions:
                    if state.is_pending():
                        total_pending += 1
                    # we don't need to check past our number_to_wait_for so exiting
                    # this loop saves on database load
                    if total_pending > number_to_resume_on:
                        break

                # check if we exited the loop above with less pending than needed
                if total_pending <= number_to_resume_on:
                    done_waiting = True  # exits while loop

                # sleep before checking again to save on database hits
                time.sleep(sleep_step)

        search_datatable.write_output_summary(directory)

        # ---------------------------------------------------------------------
        # Starting search
        # ---------------------------------------------------------------------

        # Now we are going to cycle through factor sizes, starting at the lowest
        # and working our way up. We do this because:
        #   1. smaller unitcells run much faster (for DFT-calculation)
        #   2. smaller compositions converge much faster (for number of
        #       calcs required)
        #   3. results from smaller unitcells can help to speed up the larger
        #       compositional searches.

        for natoms in range(1, max_atoms + 1):
            # for stage_number in range(1, 4) --> consider adding substages

            # Setting stop conditions for each search will be
            # essential in ensure we don't waste too much calculation time
            # on one composition -- while simultaniously making sure we
            # searched a compositon enough.
            min_structures_exact = int(5 * natoms)
            best_survival_cutoff = int(20 * natoms)
            max_structures = int(40 * natoms)
            convergence_cutoff = 0.01  # 10 meV as looser convergence limit

            # OPTIMIZE: This code below is for printing out the cutoff limits
            # I keep this here for early development as we figure out ideal
            # stopping conditions.
            # for n in range(1, 20):
            #     min_structures_exact = int(5 * n)
            #     best_survival_cutoff = int(20 * n)
            #     max_structures = int(40 * n)
            #     # convergence_cutoff = 0.01
            #     print(
            #         f"{n}\t{min_structures_exact}\t"
            #         f"{best_survival_cutoff}\t{max_structures}"
            #         # f"\t{convergence_cutoff}"
            #     )

            logging.info(f"Beginning compositional searches with natoms = {natoms}")
            current_compositions = [c for c in compositions if c.num_atoms == natoms]

            for composition in current_compositions:
                # OPTIMIZE: Should I just skip single-element compositions?
                # BUG: I do skip these for now because compositions like "N1"
                # are failing in vasp for me.
                if len(composition.elements) == 1:
                    logging.warn(f"Skipping single-element composition {composition}")
                    continue

                # For this first cycle, we only look at non-reduced compositions
                # up to the max_stoich_factor. This means Ca2N and max factor
                # of 4 would only look up to Ca8N4 and skip any compositions
                # with more atoms.
                current_factor = int(
                    composition.reduced_composition.num_atoms / composition.num_atoms
                )
                if current_factor > max_stoich_factor:
                    logging.warn(
                        f"Skipping composition {composition} bc it exceeds the max "
                        f"factor of {max_stoich_factor}"
                    )
                    continue

                # If we pass the checks above, we can start the new fixed-composition
                # search for this composition
                cls.fixed_comp_workflow.run(
                    composition=composition,
                    subworkflow_name=subworkflow_name,
                    subworkflow_kwargs=subworkflow_kwargs,
                    min_structures_exact=min_structures_exact,
                    best_survival_cutoff=best_survival_cutoff,
                    max_structures=max_structures,
                    directory=directory / composition.reduced_formula,
                    convergence_cutoff=convergence_cutoff,
                    nfirst_generation=nfirst_generation,
                    nsteadystate=nsteadystate,
                    sleep_step=sleep_step,
                    # Because we submitted all steady states above, we don't
                    # need the other workflows to do these anymore.
                    singleshot_sources=[],
                )

                # after each fixed-composition, we can reevaluate the hull diagram
                search_datatable.write_output_summary(directory)



# -*- coding: utf-8 -*-

from simmate.apps.badelf.models import BadElf
from simmate.apps.badelf.workflows.base import BadElfBase
from simmate.database import connect


class BadElf__Badelf__Badelf(BadElfBase):
    """
    This class runs a Bader and BadELF analysis without running a static
    energy calculation. The directory it is run in must already have an
    ELFCAR and CHGCAR with the same grid size as well as the AECCAR0 and
    AECCAR2 files.

    This workflow must be run with a specified directory that already exists!
    If you're using a settings.yaml file it should be in the directory above.
    You will also need to place a POSCAR file in the directory above.
    """

    use_database = True
    database_table = BadElf



# -*- coding: utf-8 -*-

from django.apps import AppConfig


class CoreComponentsConfig(AppConfig):
    name = "simmate.website.core_components"


class WorkflowsConfig(AppConfig):
    name = "simmate.website.workflows"


class WorkflowEngineConfig(AppConfig):
    name = "simmate.website.engine"


class DataExplorerConfig(AppConfig):
    name = "simmate.website.data_explorer"


class TestAppConfig(AppConfig):
    name = "simmate.website.test_app"



# -*- coding: utf-8 -*-

import numpy
from ase import Atoms
from ase.ga.cutandsplicepairing import CutAndSplicePairing
from ase.ga.utilities import CellBounds, closest_distances_generator
from pymatgen.io.ase import AseAtomsAdaptor

from simmate.toolkit import Structure
from simmate.toolkit.transformations.base import Transformation


class Heredity(Transformation):
    """
    This is a wrapper around the `CutAndSplicePairing` in ase.ga
    https://gitlab.com/ase/ase/-/blob/master/ase/ga/cutandsplicepairing.py
    """

    # TODO: I need to figure out how to implement more than two structures
    # might be useful for a pymatgen rewrite:
    # pymatgen.transformations.advanced_transformations.SlabTransformation

    name = "from_ase.Heredity"
    io_scale = "many_to_one"
    ninput = 2
    allow_parallel = False

    @staticmethod
    def apply_transformation(
        structures: list[Structure],
        ratio_of_covalent_radii: float = 0.1,
    ) -> Structure:
        #!!! takes two structures!

        # ----------- SETUP (consider caching as class attribute) -------------
        #!!! I assume 3D structure for now. I could do things like
        # pbc=[1,1,0] for 2D in the future though
        slab = Atoms(pbc=True)

        # the closest_distances_generator is exactly the same as an
        # element-dependent distance matrix expect ASE puts this in dictionary
        # form the function requires a list of element integers
        element_ints = [element.number for element in structures[0].composition]
        # the default of the ratio of covalent radii (0.1) is based on the
        # ASE tutorial of this function
        element_distance_matrix = closest_distances_generator(
            element_ints, ratio_of_covalent_radii
        )

        # boundry limits on the lattice. I only do angle limits for now but
        # I should introduce vector length limits
        cellbounds = CellBounds(
            bounds={
                "phi": [35, 145],
                "chi": [35, 145],
                "psi": [35, 145],
            }
        )
        #'a': [3, 50], 'b': [3, 50], 'c': [3, 50]})
        # ---------------------------------------------------------------------

        # split the structures into independent variables
        # copy because we may edit these into supercells below
        structure1, structure2 = [s.copy() for s in structures]

        # ASE requires the structures to be the same size but doesn't try
        # them to scale themselves. If we have two structures have the same
        # composition, we can just make supercells to make them the same size.
        if structure1.num_sites != structure2.num_sites:
            # get the least common multiple
            lcm = numpy.lcm(structure1.num_sites, structure2.num_sites)

            # scale each structure to that lcm. We don't care about the shape,
            # just that the sites come out the
            structure1.make_supercell([int(lcm / structure1.num_sites), 1, 1])
            structure2.make_supercell([int(lcm / structure2.num_sites), 1, 1])
            # BUG: This can give a structure with too large of a composition

        # if the structures each only have one site, the mutation will fail
        if structure1.num_sites == 1:
            print(
                "You cannot perform a heredity mutation on structures that "
                "only have one site!! You can make this a supercell and "
                "try again though."
            )
            return False
        # TODO: should I should make the supercell for them?

        # now we can make the generator
        casp = CutAndSplicePairing(
            # indicated dimensionality
            slab=slab,
            # number of atoms to optimize. I set this to all
            n_top=int(structure1.composition.num_atoms),
            # distance cutoff matrix
            blmin=element_distance_matrix,
            #!!! I understand this as number of unique vectors. Is that right?
            number_of_variable_cell_vectors=3,
            # p1=1, # probablitiy of shifting
            # p2=0.05, # probablitiy of shifting
            # minimum fraction of atoms that a parent must contribute.
            # None is one atom.
            # minfrac=None,
            cellbounds=cellbounds,
            # test_dist_to_slab=True,
            # use_tags=False,
            # rng=np.random,verbose=False
        )

        # first I need to convert the structures to an ASE atoms object
        structure1_ase = AseAtomsAdaptor.get_atoms(structure1)
        structure2_ase = AseAtomsAdaptor.get_atoms(structure2)

        #!!! Their code suggests the use of .get_new_individual() but I think
        # .cross() is what we'd like
        new_structure_ase = casp.cross(structure1_ase, structure2_ase)

        # if the mutation fails, None is returned
        if not new_structure_ase:
            return False

        # if it was successful, we have a new Atoms object
        # now convert back to a pymatgen object
        new_structure = AseAtomsAdaptor.get_structure(new_structure_ase)

        return new_structure



# -*- coding: utf-8 -*-

"""
We need to map which potential to grab for each element based off of the type
of calculation we are doing. To see how these were selected, you should look
at the following:

QE docs on potentials:
    https://www.quantum-espresso.org/pseudopotentials/
    https://pseudopotentials.quantum-espresso.org/
    
Standard solid-state pseudopotentials (SSSP):
    https://www.materialscloud.org/discover/sssp

Psuedo-Dojo:
    http://www.pseudo-dojo.org/index.html
"""

import json
import logging
import tarfile
import urllib

from simmate.configuration import settings
from simmate.utilities import get_directory


def setup_sssp() -> bool:
    """
    Downloads all of the SSSP potentials and mapping files needed for this app
    """
    logging.info("Setting up Standard solid-state pseudopotentials (SSSP).")
    for file_type in ["mappings", "psuedos"]:
        for psuedo_type in ["precision", "efficiency"]:
            _setup_sssp_single(file_type, psuedo_type)
    logging.info("Done!")


def _setup_sssp_single(file_type: str, psuedo_type: str):
    # we use get_dir to create the folder if it doesn't exist yet
    qe_directory = get_directory(settings.config_directory / "quantum_espresso")

    # determine which file we are trying to setup/download
    if file_type == "mappings" and psuedo_type == "precision":
        filename = "SSSP_1.3.0_PBE_precision.json"
    elif file_type == "psuedos" and psuedo_type == "precision":
        filename = "SSSP_1.3.0_PBE_precision.tar.gz"
    elif file_type == "mappings" and psuedo_type == "efficiency":
        filename = "SSSP_1.3.0_PBE_efficiency.json"
    elif file_type == "psuedos" and psuedo_type == "efficiency":
        filename = "SSSP_1.3.0_PBE_efficiency.tar.gz"

    # check if file is already present -- otherwise download it
    # TODO: I may need to switch these to a simmate CDN if I get blocked
    base_url = "https://archive.materialscloud.org/record/file?record_id=1732&filename="
    file = qe_directory / filename
    if not file.exists():
        logging.info(f"Downloading SSSP {psuedo_type} {file_type}...")
        urllib.request.urlretrieve(base_url + filename, file)
    else:
        logging.info(f"Found existing SSSP {psuedo_type} {file_type}.")

    # psuedos files are downloaded in compressed format (tar.gz). We need
    # to unpack these and move them into the /potentials folder
    if file_type == "psuedos":
        with tarfile.open(file, "r:gz") as tar:
            # Extract all contents to the specified directory
            tar.extractall(path=settings.quantum_espresso.psuedo_dir)


def _load_mappings(json_file: str):
    filename = settings.config_directory / "quantum_espresso" / json_file
    if not filename.exists():
        return {}
    with filename.open() as file:
        data = json.load(file)
    return data


SSSP_PBE_EFFICIENCY_MAPPINGS = _load_mappings("SSSP_1.3.0_PBE_efficiency.json")
SSSP_PBE_PRECISION_MAPPINGS = _load_mappings("SSSP_1.3.0_PBE_precision.json")


def check_psuedo_setup() -> bool:
    """
    Checks if SSSP psuedo files have been downloaded and are ready for use.

    To do this, we simply check that the directory has the expected files
    """

    if not SSSP_PBE_EFFICIENCY_MAPPINGS or not SSSP_PBE_PRECISION_MAPPINGS:
        return False

    # compile a list of ALL files that must be present in or for this check to
    # pass
    files = [
        "../SSSP_1.3.0_PBE_precision.json",
        "../SSSP_1.3.0_PBE_precision.tar.gz",
        "../SSSP_1.3.0_PBE_efficiency.tar.gz",
        "../SSSP_1.3.0_PBE_efficiency.tar.gz",
        *[
            mapping["filename"]
            for element, mapping in SSSP_PBE_EFFICIENCY_MAPPINGS.items()
        ],
        *[
            mapping["filename"]
            for element, mapping in SSSP_PBE_PRECISION_MAPPINGS.items()
        ],
    ]

    for file in files:
        full_filename = settings.quantum_espresso.psuedo_dir / file
        if not full_filename.exists():
            breakpoint()
            return False  # one missing is enough to exit
    # TODO: consider checking md5 hashes as well

    # if we made it to this point, then everything looks good!
    return True



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.engine import S3Workflow
from simmate.toolkit import Structure


class PopulationAnalysis__Bader__BaderDev(S3Workflow):
    use_database = False
    parent_workflows = ["population-analysis.vasp-bader.bader-matproj"]
    command = "bader {charge_file}"

    @classmethod
    def get_final_command(
        cls,
        charge_file: Path | str = "CHGCAR",
        partitioning_file: Path | str = None,
        atoms_to_print: list[int] = [],
        species_to_print: str = [],
        structure: Structure = None,
        **kwargs,
    ) -> str:
        """
        Builds the command to call bader

        Args:
            charge_file (str): The name of the file containing charge density
                data.

            partitioning_file (str): The name of the file to use for partitioning.

            atoms_to_print (list):
                A list of atom indices to print. Should be
                indices starting at 0.

            species_to_print : list
                The symbol of the atom type to print.

            structure : Structure
                The pymatgen structure object for the system of interest.
        """

        # start with the minimal required command
        command_final = cls.command.format(
            charge_file=charge_file,
            partitioning_file=partitioning_file,
        )

        # Now go through the extra optional kwargs bader cli
        if partitioning_file:
            command_final += f" -ref {partitioning_file}"

        if atoms_to_print and species_to_print:
            raise Exception(
                "Set either `atoms_to_print` or `species_to_print`. "
                "You cannot set both"
            )
        elif atoms_to_print:
            indices_str = " ".join([str(i + 1) for i in atoms_to_print])
            command_final += f" -p sel_atom {indices_str}"
        elif species_to_print:
            if not structure:
                raise Exception(
                    "You must provide a struture when setting `species_to_print`"
                )
                # TODO: grab the structure from the CHGCAR
            atom_indices = structure.indices_from_symbol(species_to_print)
            indices_str = indices_str = " ".join([str(i + 1) for i in atom_indices])
            command_final += f" -p sum_atom {indices_str}"
        return command_final



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.dynamics.mit import Dynamics__Vasp__Mit
from simmate.apps.vasp.inputs import Incar
from simmate.toolkit import Structure


class Dynamics__Vasp__MvlNpt(Dynamics__Vasp__Mit):
    """
    This task is a reimplementation of pymatgen's
    [MVLNPTMDSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MVLNPTMDSet).
    """

    _incar_updates = dict(
        ALGO="Fast",
        ISIF=3,
        LANGEVIN_GAMMA__smart_langevin=True,
        LANGEVIN_GAMMA_L=1,
        MDALGO=3,
        PMASS=10,
        PSTRESS=0,
        SMASS=0,
        ENCUT=450,  # pymatgen sets to 1.5 * max of all potcars
    )


def keyword_modifier_smart_langevin(
    structure: Structure,
    langevin_config: bool = True,  # not required
):
    """
    Expands LANGEVIN_GAMMA setting based on number of species present
    """
    return [10] * structure.ntypesp


Incar.add_keyword_modifier(keyword_modifier_smart_langevin)



# -*- coding: utf-8 -*-

from simmate.apps.warren_lab.workflows.badelf.prebadelf_dft import (
    StaticEnergy__Vasp__WarrenLabPrebadelfHse,
)
from simmate.apps.warren_lab.workflows.nested_dft.relaxation_static_base import (
    RelaxationStaticBase,
)
from simmate.apps.warren_lab.workflows.relaxation.pbesol import (
    Relaxation__Vasp__WarrenLabPbesol,
)


# We want to run a PBE relaxation followed by an HSE static energy calculation.
# Copying the WAVECAR will make this much faster, but this requires that a
# WAVECAR exists. However, we generally don't want all relaxation calculations
# to save the WAVECAR so we make a custom workflow here.
class Relaxation__Vasp__WarrenLabPbeWithWavecar(Relaxation__Vasp__WarrenLabPbesol):
    """
    This workflow is the same as the typical PBEsol relaxation but with the added
    tag in the INCAR for writing the WAVECAR. This is intended to be used with
    nested workflows to increase the speed of the static energy calculation
    """

    _incar_updates = dict(LWAVE=True)


class Nested__Vasp__WarrenLabRelaxationStaticPbeHse(RelaxationStaticBase):
    """
    Runs a PBEsol quality structure relaxation, an HSE quality static energy
    calculation.This method will also write the ELFCAR and CHGCAR files
    necessary for population analysis (i.e. oxidation state and electron count)
    """

    static_energy_workflow = StaticEnergy__Vasp__WarrenLabPrebadelfHse
    # We use pbesol as our default relaxation functional because it doesn't take
    # much more time than pbe and is considered to be more accurate for solids
    # (Phys. Rev. Lett. 102, 039902 (2009))
    relaxation_workflow = Relaxation__Vasp__WarrenLabPbeWithWavecar



# -*- coding: utf-8 -*-

from simmate.apps.materials_project.workflows.relaxation.mit import (
    Relaxation__Vasp__Mit,
)


class StaticEnergy__Vasp__Mit(Relaxation__Vasp__Mit):
    """
    Runs a VASP static energy calculation using MIT Project settings.

    This is identical to relaxation/mit, but just a single ionic step.
    """

    _incar_updates = dict(
        ALGO="Normal",
        IBRION=-1,  # (optional) locks everything between ionic steps
        NSW=0,
        KSPACING=0.4,  # was 0.5 # !!! This is where we are different from pymatgen
    )



# -*- coding: utf-8 -*-

from pathlib import Path

from pandas import DataFrame

from simmate.apps.bader.outputs import ACF
from simmate.database.base_data_types import StaticEnergy, table_column


class PopulationAnalysis(StaticEnergy):
    """
    This table combines results from a static energy calculation and the follow-up
    oxidation analysis on the charge density.
    """

    class Meta:
        app_label = "workflows"

    exclude_from_summary = [
        "oxidation_states",
        "charges",
        "min_dists",
        "atomic_volumes",
        "element_list",
    ]

    oxidation_states = table_column.JSONField(blank=True, null=True)
    """
    A list of calculated oxidation states for each site.
    """

    charges = table_column.JSONField(blank=True, null=True)
    """
    A list of total "valence" electron counts for each site.
    
    WARNING: this count is dependent on the potentials used. For example, 
    Yttrium could have used a potential where 2 or even 10 electrons are used 
    as the basis for the calculation. Use 'oxidation_states' for a more 
    consistent and accurate count of valence electrons
    """

    min_dists = table_column.JSONField(blank=True, null=True)
    """
    A list of minimum radii distance for bader volumes. i.e. the minimum
    distance from the origin of the site to the bader surface. This can be used
    as a minimum radius for the site.
    """

    atomic_volumes = table_column.JSONField(blank=True, null=True)
    """
    A list of site volumes from the oxidation analysis (i.e. the bader volume)
    """

    element_list = table_column.JSONField(blank=True, null=True)
    """
    A list of all element species in order that appear in the structure.
    
    This information is stored in the 'structure' column as well, but it is 
    stored here as an extra for convenience.
    """

    vacuum_charge = table_column.FloatField(blank=True, null=True)
    """
    Total electron count that was NOT assigned to ANY site -- and therefore
    assigned to 'vacuum'.
    
    In most cases, this value should be zero.
    """

    vacuum_volume = table_column.FloatField(blank=True, null=True)
    """
    Total volume from electron density that was NOT assigned to ANY site -- 
    and therefore assigned to 'vacuum'.
    
    In most cases, this value should be zero.
    """

    nelectrons = table_column.FloatField(blank=True, null=True)
    """
    The total number of electrons involved in the charge density partitioning.
    
    WARNING: this count is dependent on the potentials used. For example, 
    Yttrium could have used a potential where 2 or even 10 electrons are used 
    as the basis for the calculation. Use 'oxidation_states' for a more 
    consistent and accurate count of valence electrons
    """

    def write_output_summary(self, directory: Path):
        super().write_output_summary(directory)
        self.write_summary_dataframe(directory)

    @classmethod
    def from_vasp_directory(cls, directory: Path, as_dict: bool = False):
        """
        A basic workup process that reads Bader analysis results from the ACF.dat
        file and calculates the corresponding oxidation states with the existing
        POTCAR files.
        """

        # For loading the static-energy data, we can just call the parent
        # method of this class.
        energy_data = StaticEnergy.from_vasp_directory(directory, as_dict=as_dict)

        # We must then look for the bader analysis data

        # load the ACF.dat file
        dataframe, extra_data = ACF(directory)

        all_data = {
            # OPTIMIZE: consider a related table for Sites
            "oxidation_states": list(dataframe.oxidation_state.values),
            "charges": list(dataframe.charge.values),
            "min_dists": list(dataframe.min_dist.values),
            "atomic_volumes": list(dataframe.atomic_vol.values),
            "element_list": list(dataframe.element.values),
            **extra_data,
            **energy_data,
        }

        return all_data if as_dict else cls(**all_data)

    def get_summary_dataframe(self):
        df = DataFrame(
            {
                "element": self.element_list,
                "oxidation_state": self.oxidation_states,
                "charge": self.charges,
                "min_dist": self.min_dists,
                "atomic_volume": self.atomic_volumes,
            }
        )
        return df

    def write_summary_dataframe(self, directory: Path):
        df = self.get_summary_dataframe()
        filename = directory / "simmate_population_summary.csv"
        df.to_csv(filename)



# -*- coding: utf-8 -*-

from functools import cached_property
from pathlib import Path

import numpy
import xmltodict
from scipy.constants import physical_constants

from simmate.toolkit import Structure


class PwscfXml:
    def __init__(self, data: dict, check: bool = False):
        # this class is really just a single dictionary with a bunch of methods
        # attatched to it for reading/picking out certain parts of the dictionary.
        # All methods and properties are parsed "lazily" -- that is they aren't
        # loaded until they are called. Once called, they are cached as well.
        self.data = data

        # TODO: run some prechecks (or just call certain methods) to ensure
        # the dictionary is properly formatted
        if check:
            raise NotImplementedError(
                "checking pwscf.xml data for errors is not yet supported"
            )

    @classmethod
    def from_directory(cls, directory: Path | str = None, **kwargs):
        """
        Loads a `PwscfXml` instance from a output directory that was
        generated by Quantum Espresso's `pw.x` program.

        NOTE: `pwscf.xml` is assumed to be present
        """
        if not directory:
            directory = Path.cwd()
        # TODO: this will load other files in the future (e.g. charge density, bands, etc.)
        return cls.from_file(Path(directory) / "pwscf.xml", **kwargs)

    @classmethod
    def from_file(cls, filename: Path | str = "pwscf.xml", **kwargs):
        """
        Loads a `PwscfXml` instance from an `pwscf.xml` output file that was
        generated by Quantum Espresso's `pw.x` program.
        """

        filename = Path(filename)
        with filename.open() as file:
            xml_data = file.read()
        data_dict = xmltodict.parse(xml_data)

        return cls.from_dict(data_dict, **kwargs)

    @classmethod
    def from_dict(cls, data: dict, **kwargs):
        """
        Loads a `PwscfXml` instance from an `pwscf.xml` output file that was
        generated by Quantum Espresso's `pw.x` program.
        """
        # from_dict is synonmous with the default __init__ method
        return cls(data=data)

    # -------------------------------------------------------------------------

    @cached_property
    def final_structure(self):
        struct_data = self.data["qes:espresso"]["output"]["atomic_structure"]
        site_data = struct_data["atomic_positions"]["atom"]

        # lattice
        lattice_matrix = []
        for vector in ["a1", "a2", "a3"]:
            vector_cleaned = []
            for x in struct_data["cell"][vector].split():
                vector_cleaned.append(float(x))
            lattice_matrix.append(vector_cleaned)

        # species
        species = [s["@name"] for s in site_data]

        # site coords
        site_coords = []
        for s in site_data:
            coords = [float(i) for i in s["#text"].split()]
            site_coords.append(coords)

        # convert everything to Angstroms
        conversion_factor = physical_constants["Bohr radius"][0] * 1e10
        lattice_matrix = numpy.array(lattice_matrix) * conversion_factor
        site_coords = numpy.array(site_coords) * conversion_factor

        return Structure(
            lattice=lattice_matrix,
            species=species,
            coords=site_coords,
            coords_are_cartesian=True,
        )

    @cached_property
    def final_energy(self) -> float:
        rydberg_to_ev = physical_constants["Rydberg constant times hc in eV"][0]
        energy_ry = float(self.data["qes:espresso"]["output"]["total_energy"]["etot"])

        # BUG: something is off here... These values don't match what is in the
        # pwscf.out file. Perhaps these here are per atom?
        natoms = len(self.final_structure)

        # convert to eV & total energy
        return energy_ry * rydberg_to_ev * natoms

    @cached_property
    def site_forces(self) -> numpy.array:
        force_data = self.data["qes:espresso"]["output"]["forces"]["#text"]

        site_forces = []
        sites = force_data.split("\n")
        for site in sites:
            forces = [float(i) for i in site.split()]
            site_forces.append(forces)

        site_forces = numpy.array(site_forces)
        # TODO: convert to differnt units...?

        return site_forces

    @cached_property
    def lattice_stress(self) -> numpy.array:
        stress_data = self.data["qes:espresso"]["output"]["stress"]["#text"]

        stress = []
        vectors = stress_data.split("\n")
        for vector in vectors:
            stress_vector = [float(i) for i in vector.split()]
            stress.append(stress_vector)

        stress = numpy.array(stress)
        # TODO: convert to differnt units...?

        return stress

    # -------------------------------------------------------------------------



# -*- coding: utf-8 -*-
from simmate.apps.warren_lab.workflows.static_energy.hse import (
    StaticEnergy__Vasp__WarrenLabHse,
)
from simmate.apps.warren_lab.workflows.static_energy.pbesol import (
    StaticEnergy__Vasp__WarrenLabPbesol,
)

# The key thing for bader analysis is that we need a very fine FFT mesh. Other
# than that, it's the same as a static energy calculation.
PREBADELF_INCAR_SETTINGS = dict(
    # Note that these set the FFT grid while the pre-Bader task sets the
    # fine FFT grid (e.g. useds NGX instead of NGXF)
    NGX__density_a=22,
    NGY__density_b=22,
    NGZ__density_c=22,
    LELF=True,  # Writes the ELFCAR
    NPAR=1,  # Must be set if LELF is set
    PREC="Single",  # ensures CHGCAR grid matches ELFCAR grid
    LAECHG=True,  # write core charge density to AECCAR0 and valence to AECCAR2
)


class StaticEnergy__Vasp__WarrenLabPrebadelfPbesol(StaticEnergy__Vasp__WarrenLabPbesol):
    """
    Runs a static energy calculation with a high-density FFT grid under settings
    from the Warren Lab. Results can be used for Bader analysis where
    the ELF is used as the reference instead of the CHGCAR.

    We do NOT recommend running this calculation on its own. Instead, you should
    use the full workflow, which runs this calculation AND the following bader
    analysis for you. This S3Task is only the first step of that workflow.
    """

    _incar_updates = PREBADELF_INCAR_SETTINGS


class StaticEnergy__Vasp__WarrenLabPrebadelfHse(StaticEnergy__Vasp__WarrenLabHse):
    """
    Runs a static energy calculation with a high-density FFT grid under HSE
    settings from the Warren Lab. Results can be used for Bader analysis where
    the ELF is used as the reference instead of the CHGCAR.

    We do NOT recommend running this calculation on its own. Instead, you should
    use the full workflow, which runs this calculation AND the following bader
    analysis for you. This S3Task is only the first step of that workflow.
    """

    _incar_updates = PREBADELF_INCAR_SETTINGS



from simmate.apps.materials_project.workflows.static_energy.matproj_hse import (
    StaticEnergy__Vasp__MatprojHse,
)


class StaticEnergy__Vasp__MatprojHsesol(StaticEnergy__Vasp__MatprojHse):
    """
    This task is a reimplementation of pymatgen's
    [MPHSERelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPHSERelaxSet)
    but converted to a static energy calculation. The pymatgen library doesn't
    not have an HSEsol static energy, so this is the make-shift alternative.
    """

    # tells VASP to use PBEsol functional
    _incar_updates = dict(GGA="PS")



"""
This file takes the functions in `views.py` and maps them to an actual website URL.
So... if you delete/update `views.py`, then this file should be updated accordingly.

If you ever venture to this level, we STRONGLY recommend you go through the
django tutorials first: https://docs.djangoproject.com/en/3.2/
"""

from django.urls import path

from .views import example_view

urlpatterns = [
    path(
        route="",
        view=example_view,
        name="example_app",
    ),
]

# NOTE: You can delete this file if you decide not to use it.



# -*- coding: utf-8 -*-

import numpy
from scipy.ndimage import gaussian_filter1d

from simmate.toolkit import Composition
from simmate.toolkit.validators.fingerprint.base import FingerprintValidator

from .featurizers import PartialRadialDistributionFunction as prdf


class PartialRdfFingerprint(FingerprintValidator):
    comparison_mode = "cos"

    @staticmethod
    def get_featurizer(
        composition: Composition,
        cutoff: float = 20.0,
        bin_size: float = 0.1,
        **kwargs,
    ):
        featurizer = prdf(cutoff, bin_size, **kwargs)

        # for this specific featurizer, you're supposed to use the .fit()
        # function with an example structure but really, all that does is get
        # a list of unique elements - so we can make that with a composition object
        featurizer.elements_ = numpy.array(
            [element.symbol for element in composition.elements]
        )

        return featurizer

    @staticmethod
    def format_fingerprint(fingerprint):
        # smooths the function
        new_fp = gaussian_filter1d(fingerprint, sigma=1)
        return new_fp



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .cluster_high_q_e import StaticEnergy__Vasp__ClusterHighQe
from .cluster_high_q import Relaxation__Vasp__ClusterHighQ
from .cluster_low_q import Relaxation__Vasp__ClusterLowQ
from .cluster_expansion import ClusterExpansion__Clease__BulkStructure

from .staged import Relaxation__Vasp__StagedCluster



# -*- coding: utf-8 -*-

from simmate.apps.vasp.error_handlers import (
    Brions,
    Brmix,
    Edddav,
    Eddrmm,
    Edwav,
    ElfKpar,
    Frozen,
    IncorrectShift,
    IncorrectSmearing,
    InsufficientBands,
    LargeSigma,
    LongVector,
    MeshSymmetry,
    NonConverging,
    PointGroup,
    PositiveEnergy,
    Posmap,
    Potim,
    Pricel,
    Pssyevx,
    RealOptlay,
    Rhosyg,
    RotationMatrix,
    RotationNonIntMatrix,
    SubspaceMatrix,
    SymprecNoise,
    Tetirr,
    TetrahedronMesh,
    TripleProduct,
    Unconverged,
    Walltime,
    Zbrent,
    Zheev,
    Zpotrf,
)
from simmate.apps.vasp.inputs import PBE_POTCAR_MAPPINGS
from simmate.apps.vasp.workflows.base import VaspWorkflow


class Relaxation__Vasp__MatprojScan(VaspWorkflow):
    """
    This task is a reimplementation of pymatgen's
    [MPScanRelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPScanRelaxSet).
    """

    description_doc_short = "based on pymatgen's MPScanRelaxSet"

    functional = "PBE"
    potcar_mappings = PBE_POTCAR_MAPPINGS

    _incar = dict(
        ALGO="All",
        EDIFF=1.0e-05,
        EDIFFG=-0.02,
        ENAUG=1360,
        ENCUT=680,
        IBRION=2,
        ISIF=3,
        ISPIN=2,
        LORBIT=11,
        LASPH=True,
        LREAL="Auto",
        LMIXTAU=True,
        LCHARG=True,
        LAECHG=True,
        LELF=True,
        LWAVE=False,
        LVTOT=True,
        METAGGA="R2scan",
        NELM=200,
        NSW=99,
        PREC="Accurate",
        LMAXMIX__smart_lmaxmix=True,
        MAGMOM__smart_magmom={
            "default": 0.6,  # note this is different from the VASP default
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        # WARNING -- pymatgen has built in logic to determine kspacing and
        # smearing method, where they default to metallic systems.
        KSPACING=0.22,
        ISMEAR=2,
        SIGMA=0.2,
        # The actual logic in pymatgen looks closer to...
        # multiple_keywords__smart_ismear={
        #     "metal": dict(
        #         ISMEAR=2,
        #         SIGMA=0.2,
        #         KSPACING=0.22,
        #     ),
        #     "non-metal": dict(
        #         ISMEAR=-5,
        #         SIGMA=0.05,
        #         KSPACING=0.44,  # also determined by formula
        #     ),
        # },
    )

    error_handlers = [
        TetrahedronMesh(),
        RotationMatrix(),
        Brmix(),
        Zpotrf(),
        SubspaceMatrix(),
        IncorrectShift(),
        RealOptlay(),
        Tetirr(),
        RotationNonIntMatrix(),
        LongVector(),
        TripleProduct(),
        Pricel(),
        Brions(),
        Zbrent(),
        InsufficientBands(),
        Pssyevx(),
        Eddrmm(),
        Edddav(),
        Edwav(),
        Zheev(),
        ElfKpar(),
        Rhosyg(),
        Posmap(),
        PointGroup(),
        SymprecNoise(),
        IncorrectSmearing(),
        MeshSymmetry(),
        Unconverged(),
        NonConverging(),
        Potim(),
        PositiveEnergy(),
        Frozen(),
        LargeSigma(),
        Walltime(),
    ]



# -*- coding: utf-8 -*-

import re
from pathlib import Path


class Figure:
    method_type: str = "objectmethod"  # or "classmethod"

    @classmethod
    @property
    def name(cls):
        # https://stackoverflow.com/questions/1175208/
        name = re.sub(r"(?<!^)(?=[A-Z])", "_", cls.__name__).lower()
        return name

    @staticmethod
    def get_plot(parent_class):
        raise NotImplementedError("Please implement a get_plot method for your figure")

    @classmethod
    def register_to_class(cls, parent_class):
        cls.parent = parent_class

        # TODO: consider making a decorator so I don't have to write out each
        # of these functions
        def get_plot(input_obj=None, **kwargs):
            plot = cls.get_plot(input_obj, **kwargs)
            return plot

        def view_plot(input_obj=None, **kwargs):
            return cls.view_plot(input_obj, **kwargs)

        def write_plot(input_obj=None, **kwargs):
            return cls.write_plot(input_obj, **kwargs)

        def get_html_div(input_obj=None, **kwargs):
            return cls.get_html_div(input_obj, **kwargs)

        # OPTIMIZE: is there better way to determine method type?
        # # https://stackoverflow.com/questions/19227724/
        # import inspect
        # if inspect.ismethod(cls.method) and cls.method.__self__ is cls:
        #     # method bound to the class, e.g. a classmethod

        if cls.method_type == "classmethod":
            setattr(
                parent_class,
                f"get_{cls.name}_plot",
                classmethod(get_plot),
            )
            setattr(
                parent_class,
                f"view_{cls.name}_plot",
                classmethod(view_plot),
            )
            setattr(
                parent_class,
                f"write_{cls.name}_plot",
                classmethod(write_plot),
            )
            setattr(
                parent_class,
                f"get_{cls.name}_html_div",
                classmethod(get_html_div),
            )
        elif cls.method_type == "objectmethod":
            setattr(
                parent_class,
                f"get_{cls.name}_plot",
                get_plot,
            )
            setattr(
                parent_class,
                f"view_{cls.name}_plot",
                view_plot,
            )
            setattr(
                parent_class,
                f"write_{cls.name}_plot",
                write_plot,
            )
            setattr(
                parent_class,
                f"get_{cls.name}_html_div",
                get_html_div,
            )

    @classmethod
    def view_plot(cls, parent_class, **kwargs):
        raise NotImplementedError(
            "Please implement a view_plot method for your figure or inherit from "
            "common figure types (e.g. a PlotlyFigure)"
        )

    @classmethod
    def write_plot(cls, directory: Path, **kwargs):
        raise NotImplementedError(
            "Please implement a write_plot method for your figure or inherit from "
            "common figure types (e.g. a PlotlyFigure)"
        )

    @classmethod
    def get_html_div(cls, **kwargs):
        raise NotImplementedError(
            "Please implement a get_html_div method for your figure or inherit from "
            "common figure types (e.g. a PlotlyFigure)"
        )
